from dataclasses import dataclass

import abc
import datetime
from apolo_sdk import JobStatusItem, ResourceNotFound
from types import TracebackType
from typing import (
    AbstractSet,
    AsyncIterator,
    Iterable,
    Mapping,
    Optional,
    Sequence,
    Type,
    Union,
    overload,
)
from yarl import URL

from apolo_flow.types import FullID, GitInfo, ImageStatus, TaskStatus


@dataclass(frozen=True)
class Project:
    id: str
    yaml_id: str
    owner: str
    cluster: str
    org_name: str
    project_name: str  # in platform scope, not apolo-flow


@dataclass(frozen=True)
class LiveJob:
    id: str
    yaml_id: str
    project_id: str
    multi: bool
    tags: Sequence[str]
    raw_id: Optional[str] = None


@dataclass(frozen=True)
class BakeMeta:
    git_info: Optional[GitInfo]


@dataclass(frozen=True)
class Bake:
    id: str
    project_id: str
    batch: str
    name: Optional[str]
    tags: Sequence[str]
    created_at: datetime.datetime
    meta: BakeMeta
    # prefix -> { id -> deps }
    graphs: Mapping[FullID, Mapping[FullID, AbstractSet[FullID]]]
    params: Optional[Mapping[str, str]]
    last_attempt: Optional["Attempt"]


@dataclass(frozen=True)
class ConfigsMeta:
    workspace: str
    flow_config_id: str
    project_config_id: Optional[str]
    action_config_ids: Mapping[str, str]


@dataclass(frozen=True)
class Attempt:
    id: str
    bake_id: str
    number: int
    created_at: datetime.datetime
    result: TaskStatus
    configs_meta: ConfigsMeta
    executor_id: Optional[str] = None


@dataclass(frozen=True)
class TaskStatusItem:
    when: datetime.datetime
    status: TaskStatus

    @classmethod
    def from_job_transition(cls, transition: JobStatusItem) -> "TaskStatusItem":
        return cls(
            when=transition.transition_time, status=TaskStatus(transition.status)
        )


@dataclass(frozen=True)
class Task:
    id: str
    yaml_id: FullID
    attempt_id: str
    raw_id: Optional[str]
    outputs: Optional[Mapping[str, str]]
    state: Optional[Mapping[str, str]]
    statuses: Sequence[TaskStatusItem]

    @property
    def status(self) -> TaskStatus:
        return self.statuses[-1].status

    @property
    def created_at(self) -> datetime.datetime:
        return self.statuses[0].when

    @property
    def finished_at(self) -> Optional[datetime.datetime]:
        if self.status.is_finished:
            return self.statuses[-1].when
        return None


@dataclass(frozen=True)
class ConfigFile:
    id: str
    bake_id: str
    filename: str
    content: str


@dataclass(frozen=True)
class CacheEntry:
    id: str
    project_id: str
    task_id: FullID
    batch: str
    key: str
    created_at: datetime.datetime
    outputs: Mapping[str, str]
    state: Mapping[str, str]
    raw_id: str = ""


@dataclass(frozen=True)
class BakeImage:
    id: str
    bake_id: str
    yaml_defs: Sequence[FullID]
    ref: str
    status: ImageStatus
    context_on_storage: Optional[URL] = None
    dockerfile_rel: Optional[str] = None
    builder_job_id: Optional[str] = None


class _Unset:
    pass


class Storage(abc.ABC):
    async def __aenter__(self) -> "Storage":
        return self

    async def __aexit__(
        self,
        exc_typ: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        await self.close()

    async def check_can_create_for_user_context(
        self, project_name: str, cluster_name: str, org_name: Optional[str] = None
    ) -> None:
        pass

    @abc.abstractmethod
    def with_retry_read(self) -> "Storage":
        pass

    @abc.abstractmethod
    async def close(self) -> None:
        pass

    @overload
    def project(self, *, id: str) -> "ProjectStorage":
        pass

    @overload
    def project(
        self,
        *,
        yaml_id: str,
        project_name: Optional[str] = None,
        owner: Optional[str] = None,
        cluster: Optional[str] = None,
        org_name: Optional[str] = None,
    ) -> "ProjectStorage":
        pass

    @abc.abstractmethod
    def project(
        self,
        *,
        id: Optional[str] = None,
        yaml_id: Optional[str] = None,
        project_name: Optional[str] = None,
        cluster: Optional[str] = None,
        org_name: Optional[str] = None,
        owner: Optional[str] = None,
    ) -> "ProjectStorage":
        pass

    @abc.abstractmethod
    def bake(self, *, id: str) -> "BakeStorage":
        pass

    @abc.abstractmethod
    async def create_project(
        self,
        yaml_id: str,
        project_name: str,
        cluster: Optional[str] = None,
        org_name: Optional[str] = None,
    ) -> Project:
        pass

    @abc.abstractmethod
    def list_projects(
        self,
        name: Optional[str] = None,
        cluster: Optional[str] = None,
        project_name: Optional[str] = None,
        org_name: Optional[str] = None,
    ) -> AsyncIterator[Project]:
        pass

    async def get_or_create_project(
        self,
        yaml_id: str,
        project_name: str,
        cluster: str,
        org_name: Optional[str] = None,
        owner: Optional[str] = None,
    ) -> Project:
        try:
            return await self.project(
                yaml_id=yaml_id,
                cluster=cluster,
                org_name=org_name,
                owner=owner,
                project_name=project_name,
            ).get()
        except ResourceNotFound:
            await self.check_can_create_for_user_context(
                project_name, cluster, org_name
            )
            return await self.create_project(yaml_id, project_name, cluster, org_name)


class ProjectStorage(abc.ABC):
    @abc.abstractmethod
    async def get(self) -> Project:
        pass

    @abc.abstractmethod
    async def delete(self) -> None:
        pass

    @abc.abstractmethod
    def list_bakes(
        self,
        tags: Optional[AbstractSet[str]] = None,
        since: Optional[datetime.datetime] = None,
        until: Optional[datetime.datetime] = None,
        recent_first: bool = False,
    ) -> AsyncIterator[Bake]:
        pass

    @abc.abstractmethod
    async def create_bake(
        self,
        batch: str,
        # prefix -> { id -> deps }
        meta: BakeMeta,
        graphs: Mapping[FullID, Mapping[FullID, AbstractSet[FullID]]],
        params: Optional[Mapping[str, str]] = None,
        name: Optional[str] = None,
        tags: Sequence[str] = (),
    ) -> Bake:
        pass

    @overload
    def bake(self, *, id: str) -> "BakeStorage":
        pass

    @overload
    def bake(self, *, name: str) -> "BakeStorage":
        pass

    @abc.abstractmethod
    def bake(
        self,
        *,
        id: Optional[str] = None,
        name: Optional[str] = None,
    ) -> "BakeStorage":
        pass

    @abc.abstractmethod
    async def create_cache_entry(
        self,
        task_id: FullID,
        batch: str,
        key: str,
        outputs: Mapping[str, str],
        state: Mapping[str, str],
        raw_id: str,
    ) -> CacheEntry:
        pass

    @abc.abstractmethod
    async def delete_cache_entries(
        self,
        batch: Optional[str] = None,
        task_id: Optional[FullID] = None,
    ) -> None:
        pass

    @overload
    def cache_entry(self, *, id: str) -> "CacheEntryStorage":
        pass

    @overload
    def cache_entry(
        self, *, task_id: FullID, batch: str, key: str
    ) -> "CacheEntryStorage":
        pass

    @abc.abstractmethod
    def cache_entry(
        self,
        *,
        id: Optional[str] = None,
        task_id: Optional[FullID] = None,
        batch: Optional[str] = None,
        key: Optional[str] = None,
    ) -> "CacheEntryStorage":
        pass

    @abc.abstractmethod
    def list_live_jobs(
        self,
    ) -> AsyncIterator[LiveJob]:
        pass

    @abc.abstractmethod
    async def create_live_job(
        self,
        yaml_id: str,
        multi: bool,
        tags: Iterable[str],
        raw_id: Optional[str] = None,
    ) -> LiveJob:
        pass

    @abc.abstractmethod
    async def replace_live_job(
        self,
        yaml_id: str,
        multi: bool,
        tags: Iterable[str],
        raw_id: Optional[str] = None,
    ) -> LiveJob:
        pass

    @overload
    def live_job(self, *, id: str) -> "LiveJobStorage":
        pass

    @overload
    def live_job(
        self,
        *,
        yaml_id: str,
    ) -> "LiveJobStorage":
        pass

    @abc.abstractmethod
    def live_job(
        self,
        *,
        id: Optional[str] = None,
        yaml_id: Optional[str] = None,
    ) -> "LiveJobStorage":
        pass


class BakeStorage(abc.ABC):
    @abc.abstractmethod
    async def get(self) -> Bake:
        pass

    @abc.abstractmethod
    def list_attempts(self) -> AsyncIterator[Attempt]:
        pass

    @abc.abstractmethod
    async def create_attempt(
        self,
        configs_meta: ConfigsMeta,
        number: Optional[int] = None,
        executor_id: Optional[str] = None,
        result: TaskStatus = TaskStatus.PENDING,
    ) -> Attempt:
        pass

    @abc.abstractmethod
    async def create_config_file(
        self,
        filename: str,
        content: str,
    ) -> ConfigFile:
        pass

    @abc.abstractmethod
    def list_bake_images(self) -> AsyncIterator[BakeImage]:
        pass

    @abc.abstractmethod
    async def create_bake_image(
        self,
        yaml_defs: Sequence[FullID],
        ref: str,
        status: ImageStatus = ImageStatus.PENDING,
        context_on_storage: Optional[URL] = None,
        dockerfile_rel: Optional[str] = None,
        builder_job_id: Optional[str] = None,
    ) -> BakeImage:
        pass

    @overload
    def attempt(self, *, id: str) -> "AttemptStorage":
        pass

    @overload
    def attempt(self, *, number: int) -> "AttemptStorage":
        pass

    @abc.abstractmethod
    def attempt(
        self,
        *,
        id: Optional[str] = None,
        number: Optional[int] = None,
    ) -> "AttemptStorage":
        pass

    def last_attempt(
        self,
    ) -> "AttemptStorage":
        return self.attempt(number=-1)

    @abc.abstractmethod
    def config_file(
        self,
        *,
        id: str,
    ) -> "ConfigFileStorage":
        pass

    @overload
    def bake_image(self, *, id: str) -> "BakeImageStorage":
        pass

    @overload
    def bake_image(self, *, ref: str) -> "BakeImageStorage":
        pass

    @abc.abstractmethod
    def bake_image(
        self,
        *,
        id: Optional[str] = None,
        ref: Optional[str] = None,
    ) -> "BakeImageStorage":
        pass


class AttemptStorage(abc.ABC):
    @abc.abstractmethod
    async def get(self) -> Attempt:
        pass

    @abc.abstractmethod
    async def update(
        self,
        *,
        executor_id: Union[Optional[str], Type[_Unset]] = _Unset,
        result: Union[TaskStatus, Type[_Unset]] = _Unset,
    ) -> Attempt:
        pass

    @abc.abstractmethod
    def list_tasks(self) -> AsyncIterator[Task]:
        pass

    @abc.abstractmethod
    async def create_task(
        self,
        yaml_id: FullID,
        raw_id: Optional[str],
        status: Union[TaskStatusItem, TaskStatus],
        outputs: Optional[Mapping[str, str]] = None,
        state: Optional[Mapping[str, str]] = None,
    ) -> Task:
        pass

    @overload
    def task(self, *, id: str) -> "TaskStorage":
        pass

    @overload
    def task(self, *, yaml_id: FullID) -> "TaskStorage":
        pass

    @abc.abstractmethod
    def task(
        self,
        *,
        id: Optional[str] = None,
        yaml_id: Optional[FullID] = None,
    ) -> "TaskStorage":
        pass


class TaskStorage(abc.ABC):
    @abc.abstractmethod
    async def get(self) -> Task:
        pass

    @abc.abstractmethod
    async def update(
        self,
        *,
        outputs: Union[Optional[Mapping[str, str]], Type[_Unset]] = _Unset,
        state: Union[Optional[Mapping[str, str]], Type[_Unset]] = _Unset,
        new_status: Optional[Union[TaskStatusItem, TaskStatus]] = None,
    ) -> Task:
        pass


class ConfigFileStorage(abc.ABC):
    @abc.abstractmethod
    async def get(self) -> ConfigFile:
        pass


class CacheEntryStorage(abc.ABC):
    @abc.abstractmethod
    async def get(self) -> CacheEntry:
        pass


class BakeImageStorage(abc.ABC):
    @abc.abstractmethod
    async def get(self) -> BakeImage:
        pass

    @abc.abstractmethod
    async def update(
        self,
        *,
        status: Union[ImageStatus, Type[_Unset]] = _Unset,
        builder_job_id: Union[Optional[str], Type[_Unset]] = _Unset,
    ) -> BakeImage:
        pass


class LiveJobStorage(abc.ABC):
    @abc.abstractmethod
    async def get(self) -> LiveJob:
        pass



import dataclasses

import apolo_extras
import click
import datetime
from apolo_cli import __version__ as cli_version
from apolo_sdk import Client, ResourceNotFound, __version__ as sdk_version
from collections import defaultdict
from graphviz import Digraph
from operator import attrgetter
from rich import box
from rich.console import Console
from rich.panel import Panel
from rich.table import Table
from types import TracebackType
from typing import (
    AbstractSet,
    AsyncContextManager,
    AsyncIterator,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    Union,
    cast,
)
from yarl import URL

import apolo_flow

from . import ast
from .batch_executor import BatchExecutor, LocalsBatchExecutor, get_running_flow
from .colored_topo_sorter import ColoredTopoSorter
from .commands import CmdProcessor
from .config_loader import BatchLocalCL
from .context import (
    EMPTY_ROOT,
    EarlyBatch,
    EarlyLocalCall,
    ProjectCtx,
    RunningBatchFlow,
    setup_project_ctx,
)
from .expr import EvalError, MultiError
from .parser import ConfigDir
from .storage.base import (
    Attempt,
    Bake,
    BakeImage,
    BakeMeta,
    BakeStorage,
    ProjectStorage,
    Storage,
)
from .types import FullID, LocalPath, TaskStatus
from .utils import (
    CommandRunner,
    GlobalOptions,
    collect_git_info,
    encode_global_options,
    fmt_datetime,
    fmt_timedelta,
    make_cmd_exec,
)


EXECUTOR_IMAGE = f"ghcr.io/neuro-inc/apolo-flow:{apolo_flow.__version__}"


GRAPH_COLORS = {
    TaskStatus.PENDING: "skyblue",
    TaskStatus.RUNNING: "steelblue",
    TaskStatus.SUCCEEDED: "limegreen",
    TaskStatus.CANCELLED: "orange",
    TaskStatus.SKIPPED: "magenta",
    TaskStatus.CACHED: "yellowgreen",
    TaskStatus.FAILED: "orangered",
    TaskStatus.UNKNOWN: "crimson",
}


class BakeFailedError(Exception):
    def __init__(self, status: TaskStatus):
        self.status = status


async def iter_flows(top_flow: EarlyBatch) -> AsyncIterator[Tuple[FullID, EarlyBatch]]:
    to_check: List[Tuple[FullID, EarlyBatch]] = [((), top_flow)]
    while to_check:
        prefix, flow = to_check.pop(0)
        yield prefix, flow
        for tid in flow.graph:
            if await flow.is_action(tid):
                sub_flow = await flow.get_action_early(tid)
                to_check.append((prefix + (tid,), sub_flow))


async def check_no_cycles(top_flow: EarlyBatch) -> None:
    async for _, flow in iter_flows(top_flow):
        ColoredTopoSorter(flow.graph)


async def check_local_deps(top_flow: EarlyBatch) -> None:
    # This methods works in O(kn^3), where:
    # - n is number of tasks in the flow
    # - k is maximal depths of actions
    # This complexity is because:
    # For each task (n) for task's each dependency (n) and for each remote task (n)
    # do prefix check (k). Note that task are ofter have a few dependencies,
    # so in real cases one of those n effectively const.
    #
    # If performance becomes a problem, it can be replaced
    # with Trie (prefix tree) to reduce time complexity to O(kn^2)
    # (for each task (n) for each task's dependency (n) do Trie check (k))

    runs_on_remote: Set[FullID] = set()

    async for prefix, flow in iter_flows(top_flow):
        runs_on_remote.update(
            {prefix + (tid,) for tid in flow.graph if await flow.is_task(tid)}
        )

    def _is_prefix(item: FullID, prefix: FullID) -> bool:
        if len(item) < len(prefix):
            return False
        return all(x == y for (x, y) in zip(item, prefix))

    def _remote_deps(prefix: FullID, deps: Iterable[str]) -> Iterable[FullID]:
        return (
            remote
            for dep in deps
            for remote in runs_on_remote
            if _is_prefix(remote, prefix + (dep,))
        )

    async for prefix, flow in iter_flows(top_flow):
        early_locals = cast(
            AsyncIterator[EarlyLocalCall],
            (
                await flow.get_local_early(tid)
                for tid in flow.graph
                if await flow.is_local(tid)
            ),
        )
        with_bad_deps = (
            (early_local, remote)
            async for early_local in early_locals
            for remote in _remote_deps(prefix, early_local.needs)
        )
        async for early_local, remote in with_bad_deps:
            early_local_str = ".".join(prefix + (early_local.real_id,))
            remote_str = ".".join(remote)
            raise Exception(
                f"Local action '{early_local_str}' depends on remote "
                f"task '{remote_str}'. This is not supported because "
                "all local action should succeed before "
                "remote executor starts."
            )


async def check_expressions(top_flow: RunningBatchFlow) -> None:
    errors: List[EvalError] = []
    async for _, flow in iter_flows(top_flow):
        errors += flow.validate_expressions()
    if errors:
        raise MultiError(errors)


class ImageRefNotUniqueError(Exception):
    @dataclasses.dataclass
    class ImageInfo:
        context: Optional[Union[URL, LocalPath]]
        dockerfile: Optional[Union[URL, LocalPath]]
        ast: ast.Image

    def __init__(self, ref: str, images: Sequence[ImageInfo]) -> None:
        self._ref = ref
        self._images = images

    def __str__(self) -> str:
        return (
            f"Image with ref '{self._ref}' defined multiple times "
            f"with different attributes:\n"
            + "\n".join(
                f"at {EvalError.format_pos(image.ast._start)} with params:\n"
                f"  context: {image.context or '<empty>'}\n"
                f"  dockerfile: {image.dockerfile or '<empty>'}"
                for image in self._images
            )
        )


async def check_image_refs_unique(top_flow: RunningBatchFlow) -> None:
    _tmp: Dict[str, List[ImageRefNotUniqueError.ImageInfo]] = defaultdict(list)

    async for _, flow in iter_flows(top_flow):
        for image in flow.early_images.values():
            if image.ref.startswith("image:"):
                _tmp[image.ref].append(
                    ImageRefNotUniqueError.ImageInfo(
                        context=image.context,
                        dockerfile=image.dockerfile,
                        ast=flow.get_image_ast(image.id),
                    )
                )

    errors = []
    for ref, images in _tmp.items():
        # checks rendered paths are the same (for local actions)
        contexts_differ = len({it.context for it in images}) > 1
        dockerfiles_differ = len({it.dockerfile for it in images}) > 1
        if contexts_differ or dockerfiles_differ:
            # if rendered paths do not match, we might be dealing with github action
            # and should compare ast defs
            ast_contexts_differ = len({it.ast.context for it in images}) > 1
            ast_dockerfiles_differ = len({it.ast.dockerfile for it in images}) > 1
            if ast_contexts_differ or ast_dockerfiles_differ:
                errors.append(ImageRefNotUniqueError(ref, images))
    if errors:
        raise MultiError(errors)


async def build_graphs(
    top_flow: RunningBatchFlow,
) -> Mapping[FullID, Mapping[FullID, AbstractSet[FullID]]]:
    graphs = {}

    async for prefix, flow in iter_flows(top_flow):
        graphs[prefix] = {
            prefix + (key,): {prefix + (node,) for node in nodes}
            for key, nodes in flow.graph.items()
        }

    return graphs


async def upload_image_data(
    top_flow: RunningBatchFlow,
    apolo_runner: CommandRunner,
    storage: BakeStorage,
) -> List[BakeImage]:
    @dataclasses.dataclass
    class _TmpData:
        context_on_storage: Optional[URL]
        dockerfile_rel: Optional[str]
        yaml_defs: List[FullID]

    _tmp: Dict[str, _TmpData] = {}

    async for prefix, flow in iter_flows(top_flow):
        for image in flow.early_images.values():
            if isinstance(image.context, LocalPath):
                # Reusing image ref between bakes introduces
                # race condition anyway, so we can safely use it
                # as remote context dir name
                storage_context_dir: Optional[URL] = image.get_ctx_storage_dir(
                    top_flow.project_name, top_flow.project_id
                )
            else:
                storage_context_dir = image.context

            dockerfile_rel = None
            if image.dockerfile_rel:
                dockerfile_rel = str(image.dockerfile_rel.as_posix())

            prev_entry = _tmp.get(image.ref)
            if prev_entry is not None:
                # Validation is done before
                prev_entry.yaml_defs.append(prefix + (image.id,))
            else:
                if isinstance(image.context, LocalPath):
                    await apolo_runner(
                        "mkdir",
                        "--parents",
                        str(storage_context_dir),
                    )
                    await apolo_runner(
                        "cp",
                        "--recursive",
                        "--update",
                        "--no-target-directory",
                        str(image.context),
                        str(storage_context_dir),
                    )
                _tmp[image.ref] = _TmpData(
                    yaml_defs=[prefix + (image.id,)],
                    context_on_storage=storage_context_dir,
                    dockerfile_rel=dockerfile_rel,
                )

    return [
        await storage.create_bake_image(
            ref=ref,
            yaml_defs=entry.yaml_defs,
            context_on_storage=entry.context_on_storage,
            dockerfile_rel=entry.dockerfile_rel,
        )
        for ref, entry in _tmp.items()
    ]


class BatchRunner(AsyncContextManager["BatchRunner"]):
    def __init__(
        self,
        config_dir: ConfigDir,
        console: Console,
        client: Client,
        storage: Storage,
        global_options: GlobalOptions,
        run_apolo_cli: Optional[CommandRunner] = None,
    ) -> None:
        self._config_dir = config_dir
        self._console = console
        self._client = client
        self._storage = storage
        self._project_storage: Optional[ProjectStorage] = None
        self._config_loader: Optional[BatchLocalCL] = None
        self._project: Optional[ProjectCtx] = None
        self._run_apolo_cli = run_apolo_cli or make_cmd_exec(
            "apolo", global_options=encode_global_options(global_options)
        )
        self._global_options = global_options

    @property
    def project_id(self) -> str:
        assert self._project is not None
        return self._project.id

    @property
    def project_role(self) -> Optional[str]:
        assert self._project is not None
        return self._project.role

    @property
    def project_name(self) -> str:
        assert self._project is not None
        return self._project.project_name

    @property
    def config_loader(self) -> BatchLocalCL:
        assert self._config_loader is not None
        return self._config_loader

    @property
    def storage(self) -> ProjectStorage:
        assert self._project_storage is not None
        return self._project_storage

    async def close(self) -> None:
        if self._config_loader is not None:
            await self._config_loader.close()

    async def __aenter__(self) -> "BatchRunner":
        self._config_loader = BatchLocalCL(self._config_dir, self._client)
        self._project = await setup_project_ctx(EMPTY_ROOT, self._config_loader)
        project_name = (
            self._project.project_name or self._client.config.project_name_or_raise
        )
        project = await self._storage.get_or_create_project(
            self._project.id,
            owner=self._project.owner,
            project_name=project_name,
            cluster=self._client.config.cluster_name,
            org_name=self._client.config.org_name,
        )
        self._project_storage = self._storage.project(id=project.id)
        return self

    async def __aexit__(
        self,
        exc_typ: Optional[Type[BaseException]],
        exc_val: Optional[BaseException],
        exc_tb: Optional[TracebackType],
    ) -> None:
        await self.close()

    # Next function is also used in tests:
    async def _setup_bake(
        self,
        batch_name: str,
        params: Optional[Mapping[str, str]] = None,
        name: Optional[str] = None,
        tags: Sequence[str] = (),
    ) -> Tuple[Bake, RunningBatchFlow]:
        # batch_name is a name of yaml config inside self._workspace / .neuro
        # folder without the file extension
        self._console.log(f"[bright_black]apolo_sdk=={sdk_version}")
        self._console.log(f"[bright_black]apolo_cli=={cli_version}")
        self._console.log(f"[bright_black]apolo-extras=={apolo_extras.__version__}")
        self._console.log(f"[bright_black]apolo-flow=={apolo_flow.__version__}")
        self._console.log(f"Use config file {self.config_loader.flow_path(batch_name)}")

        # Check that the yaml is parseable
        flow = await RunningBatchFlow.create(
            self.config_loader, batch_name, "fake-bake-id", params
        )

        for volume in flow.volumes.values():
            if volume.local is not None:
                # TODO: sync volumes if needed
                raise NotImplementedError("Volumes sync is not supported")

        await check_no_cycles(flow)
        await check_local_deps(flow)
        await check_expressions(flow)
        await check_image_refs_unique(flow)
        graphs = await build_graphs(flow)

        self._console.log(
            "Check config... [green]ok[/green]",
        )

        self._console.log("Create bake...")

        bake = await self.storage.create_bake(
            batch=batch_name,
            graphs=graphs,
            params=flow.params,
            name=name,
            tags=tags,
            meta=BakeMeta(
                git_info=await collect_git_info(),
            ),
        )
        bake_storage = self.storage.bake(id=bake.id)
        config_meta = await self.config_loader.collect_configs(batch_name, bake_storage)
        await bake_storage.create_attempt(number=1, configs_meta=config_meta)

        self._console.log(
            f"Bake [b]{bake.name or bake.id}[/b] of "
            f"flow [b]{self.project_id}[/b] is created"
        )
        self._console.log("Uploading image contexts/dockerfiles...")
        await upload_image_data(flow, self._run_apolo_cli, bake_storage)

        return bake, flow

    # Next function is also used in tests:
    async def _run_locals(
        self,
        bake_id: str,
    ) -> TaskStatus:
        async with LocalsBatchExecutor.create(
            self._console,
            bake_id,
            self._client,
            self._storage,
            project_role=self.project_role,
        ) as executor:
            return await executor.run()

    async def bake(
        self,
        batch_name: str,
        local_executor: bool = False,
        params: Optional[Mapping[str, str]] = None,
        name: Optional[str] = None,
        tags: Sequence[str] = (),
    ) -> None:
        self._console.print(
            Panel(f"[bright_blue]Bake [b]{batch_name}[/b]", padding=1),
            justify="center",
        )
        bake, flow = await self._setup_bake(batch_name, params, name, tags)
        await self._run_bake(bake, flow, local_executor)

    async def _run_bake(
        self,
        bake: Bake,
        flow: RunningBatchFlow,
        local_executor: bool,
    ) -> None:
        self._console.rule("Run local actions")
        locals_result = await self._run_locals(bake.id)
        if locals_result != TaskStatus.SUCCEEDED:
            return
        self._console.rule("Run main actions")
        if local_executor:
            self._console.log(f"[bright_black]Using local executor")
            await self.process(bake.id)
        else:
            self._console.log(f"[bright_black]Starting remote executor")
            if flow.life_span:
                life_span = fmt_timedelta(flow.life_span)
            else:
                life_span = "10d"
            # TODO: Update tags `project` -> `flow` and `flow` -> `batch`
            # After performing the corresponding changes in web app upfront.
            run_args = [
                "run",
                "--pass-config",
                f"--volume=storage:.flow/logs/{bake.id}/:/root/.neuro/logs",
                f"--life-span={life_span}",
                f"--tag=project:{self.project_id}",
                f"--tag=flow:{bake.batch}",
                f"--tag=bake_id:{bake.id}",
                f"--tag=remote_executor",
                f"--project={flow.project_name}",
            ]
            project_role = self.project_role
            if project_role is not None:
                run_args.append(f"--share={project_role}")
            run_args += [
                EXECUTOR_IMAGE,
                "--",
                "apolo-flow",
                *encode_global_options(self._global_options),
                "--fake-workspace",
                "execute",
                bake.id,
            ]
            await self._run_apolo_cli(*run_args)

    async def process(
        self,
        bake_id: str,
    ) -> None:
        async with BatchExecutor.create(
            self._console,
            bake_id,
            self._client,
            self._storage,
            project_role=self.project_role,
        ) as executor:
            status = await executor.run()
            if status != TaskStatus.SUCCEEDED:
                raise BakeFailedError(status)

    def get_bakes(self) -> AsyncIterator[Bake]:
        return self.storage.list_bakes()

    async def get_bake_attempt(self, bake_id: str, *, attempt_no: int = -1) -> Attempt:
        return await self._storage.bake(id=bake_id).attempt(number=attempt_no).get()

    async def list_bakes(
        self,
        tags: AbstractSet[str] = frozenset(),
        since: Optional[datetime.datetime] = None,
        until: Optional[datetime.datetime] = None,
        recent_first: bool = False,
    ) -> None:
        def _setup_table() -> Table:
            table = Table(box=box.MINIMAL_HEAVY_HEAD)
            table.add_column(
                "ID",
                style="bold",
                width=len("bake-f6bd815b-3a3b-4ea1-b5ec-e8ab13678e3e"),
            )
            table.add_column("NAME", min_width=12)
            table.add_column("BATCH", min_width=20)
            table.add_column(
                "EXECUTOR", width=len("job-f6bd815b-3a3b-4ea1-b5ec-e8ab13678e3e")
            )
            table.add_column("STATUS", width=9)
            table.add_column("WHEN", min_width=10)
            table.show_edge = False
            return table

        header_table = _setup_table()
        self._console.print(header_table)

        async for bake in self.storage.list_bakes(
            tags=tags,
            since=since,
            until=until,
            recent_first=recent_first,
        ):
            if bake.last_attempt is None:
                self._console.print(
                    f"[yellow]Bake [b]{bake.id}[/b] is malformed, skipping"
                )
            else:
                row_table = _setup_table()
                row_table.show_header = False
                row_table.add_row(
                    bake.id,
                    bake.name or "",
                    bake.batch,
                    bake.last_attempt.executor_id or "",
                    bake.last_attempt.result,
                    fmt_datetime(bake.last_attempt.created_at),
                )
                self._console.print(row_table)

    async def inspect(
        self,
        bake_id: str,
        *,
        attempt_no: int = -1,
        output: Optional[LocalPath] = None,
        save_dot: bool = False,
        save_pdf: bool = False,
        view_pdf: bool = False,
    ) -> None:
        bake_storage = self.storage.bake(id=bake_id)
        try:
            bake = await bake_storage.get()
        except ResourceNotFound:
            self._console.print("[yellow]Bake not found")
            self._console.print(
                f"Please make sure that the bake ID [b]{bake_id}[/b] and "
                f"flow [b]{self.project_id}[/b] are correct."
            )
            exit(1)
            assert False, "unreachable"

        attempt_storage = bake_storage.attempt(number=attempt_no)
        attempt = await attempt_storage.get()

        self._console.print(f"[b]Bake id: {bake_id}[/b]")
        self._console.print(f"[b]Attempt #{attempt.number}[/b]", attempt.result)
        if attempt.executor_id:
            info = await self._client.jobs.status(attempt.executor_id)
            self._console.print(
                f"[b]Executor {attempt.executor_id}[/b]", TaskStatus(info.status)
            )

        task_table = Table(box=box.MINIMAL_HEAVY_HEAD)
        task_table.add_column("ID", style="bold")
        task_table.add_column("STATUS")
        task_table.add_column("RAW ID", style="bright_black")
        task_table.add_column("STARTED")
        task_table.add_column("FINISHED")

        tasks = [task async for task in attempt_storage.list_tasks()]

        for task in sorted(tasks, key=attrgetter("created_at")):
            task_table.add_row(
                ".".join(task.yaml_id),
                task.status,
                task.raw_id,
                fmt_datetime(task.created_at),
                fmt_datetime(task.finished_at),
            )

        self._console.print("Tasks:")
        self._console.print(task_table)

        image_table = Table(box=box.MINIMAL_HEAVY_HEAD)
        image_table.add_column("REF", style="bold")
        image_table.add_column("STATUS")
        image_table.add_column("BUILDER ID", style="bright_black")
        async for image in bake_storage.list_bake_images():
            image_table.add_row(
                image.ref,
                image.status,
                image.builder_job_id or "",
            )
        if image_table.rows:
            self._console.print("Images:")
            self._console.print(image_table)

        if output is None:
            output = LocalPath(f"{bake.id}_{attempt.number}").with_suffix(".gv")

        graphs = bake.graphs
        dot = Digraph(bake.batch, filename=str(output), strict=True, engine="dot")
        dot.attr(compound="true")
        dot.node_attr = {"style": "filled"}

        await self._subgraph(
            dot, graphs, (), {}, {task.yaml_id: task.status for task in tasks}
        )

        if save_dot:
            self._console.print(f"Saving file {dot.filename}")
            dot.save()
        if save_pdf:
            self._console.print(f"Rendering {dot.filename}.pdf")
            dot.render(view=view_pdf)
        elif view_pdf:
            self._console.print(f"Opening {dot.filename}.pdf")
            dot.view()

    async def _subgraph(
        self,
        dot: Digraph,
        graphs: Mapping[FullID, Mapping[FullID, AbstractSet[FullID]]],
        prefix: FullID,
        anchors: Dict[str, str],
        statuses: Dict[FullID, TaskStatus],
    ) -> None:
        lhead: Optional[str]
        ltail: Optional[str]
        color: Optional[str]
        first = True
        graph = graphs[prefix]
        for task_id, deps in graph.items():
            tgt = ".".join(task_id)
            name = task_id[-1]

            if first:
                anchors[".".join(prefix)] = tgt
                first = False

            if task_id in statuses:
                color = GRAPH_COLORS.get(statuses[task_id])
            else:
                color = None

            if task_id in graphs:
                lhead = "cluster_" + tgt
                with dot.subgraph(name=lhead) as subdot:
                    subdot.attr(label=f"{name}")
                    subdot.attr(compound="true")
                    subdot.attr(color=color)
                    await self._subgraph(
                        subdot,
                        graphs,
                        task_id,
                        anchors,
                        statuses,
                    )
                tgt = anchors[tgt]
            else:
                dot.node(tgt, name, color=color)
                lhead = None

            for dep in deps:
                src = ".".join(dep)
                if src in anchors:
                    # src is a subgraph
                    ltail = "cluster_" + src
                    src = anchors[src]
                else:
                    ltail = None
                dot.edge(src, tgt, ltail=ltail, lhead=lhead)

    async def logs(
        self, bake_id: str, task_id: str, *, attempt_no: int = -1, raw: bool = False
    ) -> None:
        attempt_storage = self.storage.bake(id=bake_id).attempt(number=attempt_no)
        attempt = await attempt_storage.get()
        full_id = tuple(task_id.split("."))
        try:
            task = await attempt_storage.task(yaml_id=full_id).get()
        except ResourceNotFound:
            raise click.BadArgumentUsage(f"Unknown task {task_id}")

        if not task.status.is_finished:
            raise click.BadArgumentUsage(f"Task {task_id} is not finished")

        self._console.print(f"[b]Attempt #{attempt.number}[/b]", attempt.result)
        self._console.print(f"Task [b]{task_id}[/b]", task.status)

        if not task.raw_id:
            return

        if raw:
            async for chunk in self._client.jobs.monitor(task.raw_id):
                self._console.print(chunk.decode("utf-8", "replace"), end="")
        else:
            async with CmdProcessor() as proc:
                async for chunk in self._client.jobs.monitor(task.raw_id):
                    async for line in proc.feed_chunk(chunk):
                        self._console.print(line.decode("utf-8", "replace"), end="")
                async for line in proc.feed_eof():
                    self._console.print(line.decode("utf-8", "replace"), end="")

    async def cancel(self, bake_id: str, *, attempt_no: int = -1) -> None:
        attempt_storage = self.storage.bake(id=bake_id).attempt(number=attempt_no)
        attempt = await attempt_storage.get()
        if attempt.result.is_finished:
            raise click.BadArgumentUsage(
                f"Attempt #{attempt.number} of {attempt.bake_id} is already stopped."
            )
        await attempt_storage.update(result=TaskStatus.CANCELLED)
        self._console.print(
            f"[b]Attempt #{attempt.number}[/b] of bake "
            f"[b]{attempt.bake_id}[/b] was cancelled."
        )

    async def clear_cache(
        self, batch: Optional[str] = None, task_id: Optional[str] = None
    ) -> None:
        full_id: Optional[FullID] = None
        if task_id:
            full_id = tuple(task_id.split("."))
        await self.storage.delete_cache_entries(batch, full_id)

    async def restart(
        self,
        bake_id: str,
        *,
        attempt_no: int = -1,
        from_failed: bool = True,
        local_executor: bool = False,
    ) -> None:
        bake, flow = await self._restart(
            bake_id, attempt_no=attempt_no, from_failed=from_failed
        )
        await self._run_bake(bake, flow, local_executor)

    async def _restart(
        self,
        bake_id: str,
        *,
        attempt_no: int = -1,
        from_failed: bool = True,
    ) -> Tuple[Bake, RunningBatchFlow]:
        bake_storage = self.storage.bake(id=bake_id)
        bake = await bake_storage.get()
        if bake.last_attempt and attempt_no == -1:
            last_attempt = attempt = bake.last_attempt
        else:
            attempt = await bake_storage.attempt(number=attempt_no).get()
            last_attempt = await bake_storage.last_attempt().get()
        if not attempt.result.is_finished:
            raise click.BadArgumentUsage(
                f"Cannot re-run still running attempt #{attempt.number} "
                f"of {bake.id}."
            )
        if not last_attempt.result.is_finished:
            raise click.BadArgumentUsage(
                f"Cannot re-run bake when last attempt #{last_attempt.number} "
                f"of {bake.id} is still running."
            )
        if attempt.result == TaskStatus.SUCCEEDED and from_failed:
            raise click.BadArgumentUsage(
                f"Cannot re-run successful attempt #{attempt.number} "
                f"of {bake.id} with `--from-failed` flag set.\n"
                "Hint: Try adding --no-from-failed to restart bake from the beginning."
            )
        if attempt.number >= 99:
            raise click.BadArgumentUsage(
                f"Cannot re-run {bake.id}, the number of attempts exceeded."
            )
        new_attempt = await bake_storage.create_attempt(
            number=last_attempt.number + 1,
            configs_meta=attempt.configs_meta,
        )
        if from_failed:
            new_attempt_storage = bake_storage.attempt(id=new_attempt.id)
            graphs = bake.graphs
            handled = set()  # a set of succesfully finished and not cached tasks
            tasks = {
                task.yaml_id: task
                async for task in bake_storage.attempt(id=attempt.id).list_tasks()
            }
            for task in sorted(tasks.values(), key=attrgetter("created_at")):
                if task.status == TaskStatus.SUCCEEDED:
                    # should check deps to don't process post-actions with
                    # always() precondition
                    prefix = task.yaml_id[:-1]
                    graph = graphs[prefix]
                    deps = graph[task.yaml_id]
                    if not deps or all(dep in handled for dep in deps):
                        if (
                            prefix in tasks
                            and tasks[prefix].status != TaskStatus.SUCCEEDED
                        ):
                            # If action didn't succeeded, we should create task manually
                            await new_attempt_storage.create_task(
                                yaml_id=prefix,
                                status=TaskStatus.PENDING,
                                raw_id=None,
                            )
                        # TODO allow to create task with multiple statuses
                        # and copy them from old task
                        await new_attempt_storage.create_task(
                            yaml_id=task.yaml_id,
                            status=TaskStatus.SUCCEEDED,
                            raw_id=task.raw_id,
                            outputs=task.outputs,
                            state=task.state,
                        )
                        handled.add(task.id)

        self._console.print(f"[b]Attempt #{new_attempt.number}[/b] is created")
        flow = await get_running_flow(
            bake, self._client, bake_storage, new_attempt.configs_meta
        )

        return bake, flow



from apolo_sdk import PluginManager


APOLO_FLOW_UPGRADE = """\
You are using Apolo Flow {old_ver}, however {new_ver} is available.
You should consider upgrading via the following command:
    python -m pip install --upgrade apolo-flow
"""


def get_apolo_flow_txt(old: str, new: str) -> str:
    return APOLO_FLOW_UPGRADE.format(old_ver=old, new_ver=new)


def setup(manager: PluginManager) -> None:
    manager.version_checker.register("apolo-flow", get_apolo_flow_txt)



from dataclasses import replace

import abc
import aiohttp
import datetime
import json
from apolo_sdk import Client, Project as ApoloProject
from contextlib import asynccontextmanager
from typing import (
    AbstractSet,
    Any,
    AsyncIterator,
    Callable,
    Dict,
    Generic,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)
from typing_extensions import TypedDict
from yarl import URL

from apolo_flow.storage.base import (
    Attempt,
    AttemptStorage,
    Bake,
    BakeImage,
    BakeImageStorage,
    BakeMeta,
    BakeStorage,
    CacheEntry,
    CacheEntryStorage,
    ConfigFile,
    ConfigFileStorage,
    ConfigsMeta,
    LiveJob,
    LiveJobStorage,
    Project,
    ProjectStorage,
    Storage,
    Task,
    TaskStatusItem,
    TaskStorage,
    _Unset,
)
from apolo_flow.types import FullID, GitInfo, ImageStatus, TaskStatus
from apolo_flow.utils import RetryConfig, retry


def _id_from_json(sid: str) -> FullID:
    if not sid:
        return ()
    else:
        return tuple(sid.split("."))


def _id_to_json(full_id: FullID) -> str:
    return ".".join(full_id)


def _now() -> datetime.datetime:
    return datetime.datetime.now(datetime.timezone.utc)


def _dt2str(dt: datetime.datetime) -> str:
    return dt.isoformat(timespec="seconds")


def _parse_project_payload(data: Mapping[str, Any]) -> Project:
    return Project(
        id=data["id"],
        yaml_id=data["name"],
        owner=data["owner"],
        cluster=data["cluster"],
        org_name=data["org_name"],
        project_name=data["project_name"],
    )


def _parse_live_job_payload(data: Mapping[str, Any]) -> LiveJob:
    return LiveJob(
        id=data["id"],
        yaml_id=data["yaml_id"],
        project_id=data["project_id"],
        multi=data["multi"],
        tags=data["tags"],
        raw_id=data.get("raw_id", None),
    )


def _parse_cache_entry_payload(data: Mapping[str, Any]) -> CacheEntry:
    return CacheEntry(
        id=data["id"],
        project_id=data["project_id"],
        batch=data["batch"],
        task_id=_id_from_json(data["task_id"]),
        key=data["key"],
        created_at=datetime.datetime.fromisoformat(data["created_at"]),
        outputs=data["outputs"],
        state=data["state"],
        raw_id=data.get("raw_id", None),
    )


def _parse_task_payload(data: Mapping[str, Any]) -> Task:
    return Task(
        id=data["id"],
        yaml_id=_id_from_json(data["yaml_id"]),
        attempt_id=data["attempt_id"],
        outputs=data["outputs"],
        state=data["state"],
        raw_id=data.get("raw_id", None),
        statuses=[
            TaskStatusItem(
                when=datetime.datetime.fromisoformat(item["created_at"]),
                status=TaskStatus(item["status"]),
            )
            for item in data["statuses"]
        ],
    )


def _parse_config_file_payload(data: Mapping[str, Any]) -> ConfigFile:
    return ConfigFile(
        id=data["id"],
        bake_id=data["bake_id"],
        filename=data["filename"],
        content=data["content"],
    )


def _parse_bake_payload(data: Mapping[str, Any]) -> Bake:
    graphs = {}
    for pre, gr in data["graphs"].items():
        graphs[_id_from_json(pre)] = {
            _id_from_json(full_id): {_id_from_json(dep) for dep in deps}
            for full_id, deps in gr.items()
        }
    last_attempt: Optional[Attempt] = data.get("last_attempt")
    if last_attempt:
        last_attempt = _parse_attempt_payload(data["last_attempt"])

    git_info: Optional[GitInfo] = None
    if data["meta"].get("git_info"):
        git_info = GitInfo(
            sha=data["meta"]["git_info"]["sha"],
            branch=data["meta"]["git_info"]["branch"],
            tags=data["meta"]["git_info"]["tags"],
        )

    return Bake(
        id=data["id"],
        project_id=data["project_id"],
        name=data["name"],
        batch=data["batch"],
        tags=data["tags"],
        created_at=datetime.datetime.fromisoformat(data["created_at"]),
        graphs=graphs,
        params=data["params"],
        last_attempt=last_attempt,
        meta=BakeMeta(git_info=git_info),
    )


def _parse_attempt_payload(data: Mapping[str, Any]) -> Attempt:
    return Attempt(
        id=data["id"],
        bake_id=data["bake_id"],
        number=data["number"],
        created_at=datetime.datetime.fromisoformat(data["created_at"]),
        result=TaskStatus(data["result"]),
        configs_meta=ConfigsMeta(
            workspace=data["configs_meta"]["workspace"],
            flow_config_id=data["configs_meta"]["flow_config_id"],
            project_config_id=data["configs_meta"]["project_config_id"],
            action_config_ids=data["configs_meta"]["action_config_ids"],
        ),
        executor_id=data.get("executor_id", None),
    )


def _parse_bake_image_payload(data: Mapping[str, Any]) -> BakeImage:
    context_on_storage_raw = data.get("context_on_storage", None)
    context_on_storage: Optional[URL] = None
    if context_on_storage_raw is not None:
        context_on_storage = URL(context_on_storage_raw)
    return BakeImage(
        id=data["id"],
        bake_id=data["bake_id"],
        yaml_defs=[_id_from_json(sid) for sid in data["yaml_defs"]],
        ref=data["ref"],
        status=ImageStatus(data["status"]),
        context_on_storage=context_on_storage,
        dockerfile_rel=data.get("dockerfile_rel", None),
        builder_job_id=data.get("builder_job_id", None),
    )


class RawApiClient:
    def __init__(self, client: Client) -> None:
        self._core = client._core
        self._config = client.config
        self._base_url = client.config.api_url

    @asynccontextmanager
    async def _request(
        self,
        method: str,
        url_suffix: str,
        data: Optional[Mapping[str, Any]] = None,
        headers: Optional[Mapping[str, str]] = None,
        params: Union[Sequence[Tuple[str, str]], Mapping[str, str], None] = None,
    ) -> AsyncIterator[aiohttp.ClientResponse]:
        url = self._base_url / url_suffix
        auth = await self._config._api_auth()
        async with self._core.request(
            method,
            url,
            headers=headers,
            auth=auth,
            json=data,
            params=params,
        ) as resp:
            yield resp

    E = TypeVar("E")

    async def list(
        self,
        url_suffix: str,
        mapper: Callable[[Mapping[str, Any]], E],
        params: Union[Sequence[Tuple[str, str]], Mapping[str, str], None] = None,
    ) -> AsyncIterator[E]:
        headers = {"Accept": "application/x-ndjson"}
        async with self._request(
            "GET", url_suffix, headers=headers, params=params
        ) as resp:
            async for line in resp.content:
                yield mapper(json.loads(line))

    async def create(
        self,
        url_suffix: str,
        data: Mapping[str, Any],
        mapper: Callable[[Mapping[str, Any]], E],
    ) -> E:
        async with self._request("POST", url_suffix, data=data) as resp:
            return mapper(await resp.json())

    async def update(
        self,
        url_suffix: str,
        data: Mapping[str, Any],
        mapper: Callable[[Mapping[str, Any]], E],
    ) -> E:
        async with self._request("PUT", url_suffix, data=data) as resp:
            return mapper(await resp.json())

    async def patch(
        self,
        url_suffix: str,
        data: Mapping[str, Any],
        mapper: Callable[[Mapping[str, Any]], E],
    ) -> E:
        async with self._request("PATCH", url_suffix, data=data) as resp:
            return mapper(await resp.json())

    async def get(
        self,
        url_suffix: str,
        mapper: Callable[[Mapping[str, Any]], E],
        params: Union[Sequence[Tuple[str, str]], Mapping[str, str], None] = None,
    ) -> E:
        async with self._request("GET", url_suffix, params=params) as resp:
            return mapper(await resp.json())

    async def delete(
        self,
        url_suffix: str,
        params: Union[Sequence[Tuple[str, str]], Mapping[str, str], None] = None,
    ) -> None:
        async with self._request("DELETE", url_suffix, params=params):
            pass


class RetryingReadRawApiClient(RetryConfig, RawApiClient):
    def __init__(self, client: Client) -> None:
        super().__init__()
        super(RetryConfig, self).__init__(client)

    E = TypeVar("E")

    @retry
    async def get(
        self,
        url_suffix: str,
        mapper: Callable[[Mapping[str, Any]], E],
        params: Union[Sequence[Tuple[str, str]], Mapping[str, str], None] = None,
    ) -> E:
        return await super().get(url_suffix, mapper, params)

    @retry
    async def _list(
        self,
        url_suffix: str,
        mapper: Callable[[Mapping[str, Any]], E],
        params: Union[Sequence[Tuple[str, str]], Mapping[str, str], None] = None,
    ) -> List[E]:
        res = []
        async for item in super().list(url_suffix, mapper, params):
            res.append(item)
        return res

    async def list(
        self,
        url_suffix: str,
        mapper: Callable[[Mapping[str, Any]], E],
        params: Union[Sequence[Tuple[str, str]], Mapping[str, str], None] = None,
    ) -> AsyncIterator[E]:
        for item in await self._list(url_suffix, mapper, params):
            yield item


class ApiStorage(Storage):
    def __init__(
        self, client: Client, _raw_client: Optional[RawApiClient] = None
    ) -> None:
        self._client = client
        self._cluster_name = client.config.cluster_name
        self._org_name = client.config.org_name
        self._raw_client = RawApiClient(client) if _raw_client is None else _raw_client

    def with_retry_read(self) -> Storage:
        return ApiStorage(
            self._client, _raw_client=RetryingReadRawApiClient(self._client)
        )

    async def check_can_create_for_user_context(
        self,
        project_name: str,
        cluster_name: str,
        org_name: Optional[str] = None,
    ) -> None:
        await self._client.config.fetch()
        needed_key = ApoloProject.Key(cluster_name, org_name, project_name)
        if needed_key not in self._client.config.projects.keys():
            org_msg = f"organization '{org_name}', " if org_name else ""
            msg = (
                "Cannot create flow within your user context. Make sure you have access"
                f" to {org_msg}cluster '{cluster_name}' and project '{project_name}'"
            )
            raise ValueError(msg)

    async def close(self) -> None:
        pass

    def project(
        self,
        *,
        id: Optional[str] = None,
        project_name: Optional[str] = None,
        owner: Optional[str] = None,
        yaml_id: Optional[str] = None,
        cluster: Optional[str] = None,
        org_name: Optional[str] = None,
    ) -> "ProjectStorage":
        cluster = cluster or self._cluster_name
        org_name = org_name or self._org_name
        owner = owner or self._client.config.username
        project_name = project_name or self._client.config.project_name_or_raise

        if id:
            return ApiProjectStorage(self._raw_client, id=id)
        else:
            assert yaml_id
            return ApiProjectStorage(
                self._raw_client,
                args=dict(
                    yaml_id=yaml_id,
                    cluster=cluster,
                    org_name=org_name,
                    owner=owner,
                    project_name=project_name,
                ),
            )

    async def create_project(
        self,
        yaml_id: str,
        project_name: str,
        cluster: Optional[str] = None,
        org_name: Optional[str] = None,
    ) -> Project:
        return await self._raw_client.create(
            "flow/projects",
            data={
                "name": yaml_id,
                "project_name": project_name,
                "cluster": cluster or self._cluster_name,
                "org_name": org_name or self._org_name,
            },
            mapper=_parse_project_payload,
        )

    def list_projects(
        self,
        name: Optional[str] = None,
        cluster: Optional[str] = None,
        project_name: Optional[str] = None,
        org_name: Optional[str] = None,
    ) -> AsyncIterator[Project]:
        params = []
        if name is not None:
            params += [("name", name)]
        if cluster is not None:
            params += [("cluster", cluster)]
        if project_name is not None:
            params += [("project_name", project_name)]
        if org_name is not None:
            params += [("org_name", org_name)]
        return self._raw_client.list("flow/projects", _parse_project_payload, params)

    def bake(self, *, id: str) -> "BakeStorage":
        return ApiBakeStorage(self._raw_client, id=id)


E = TypeVar("E")
ARGS = TypeVar("ARGS")


class DeferredIdMixin(Generic[E, ARGS], abc.ABC):
    _id: Optional[str]
    _args: Optional[ARGS]

    @abc.abstractmethod
    async def _retrieve_id(self, args: ARGS) -> Tuple[str, E]:
        pass

    async def get_id(self) -> Tuple[str, Optional[E]]:
        if self._id:
            return self._id, None
        assert self._args
        id_, entity = await self._retrieve_id(self._args)
        self._id = id_
        return id_, entity


class ProjectInitArgs(TypedDict):
    yaml_id: str
    cluster: str
    org_name: Optional[str]
    owner: str
    project_name: str


class ApiProjectStorage(DeferredIdMixin[Project, ProjectInitArgs], ProjectStorage):
    @overload
    def __init__(self, raw_client: RawApiClient, *, id: str):
        pass

    @overload
    def __init__(
        self,
        raw_client: RawApiClient,
        *,
        args: ProjectInitArgs,
    ):
        pass

    def __init__(
        self,
        raw_client: RawApiClient,
        *,
        id: Optional[str] = None,
        args: Optional[ProjectInitArgs] = None,
    ):
        self._raw_client = raw_client
        self._id = id
        self._args = args

    async def _retrieve_id(self, args: ProjectInitArgs) -> Tuple[str, Project]:
        params = {
            "name": args["yaml_id"],
            "cluster": args["cluster"],
            "owner": args["owner"],
            "project_name": args["project_name"],
        }
        org_name = args.get("org_name")
        if org_name:
            params["org_name"] = org_name
        project = await self._raw_client.get(
            "flow/projects/by_name", _parse_project_payload, params=params
        )
        return project.id, project

    async def get(self) -> Project:
        project_id, project = await self.get_id()
        if project:
            return project
        return await self._raw_client.get(
            f"flow/projects/{project_id}", _parse_project_payload
        )

    async def delete(self) -> None:
        project_id, _ = await self.get_id()
        await self._raw_client.delete(f"flow/projects/{project_id}")

    async def list_bakes(
        self,
        tags: Optional[AbstractSet[str]] = None,
        since: Optional[datetime.datetime] = None,
        until: Optional[datetime.datetime] = None,
        recent_first: bool = False,
    ) -> AsyncIterator[Bake]:
        project_id, _ = await self.get_id()

        params = [("project_id", project_id), ("fetch_last_attempt", "1")]
        if tags is not None:
            params += [("tags", tag) for tag in tags]
        if since is not None:
            params += [("since", since.isoformat())]
        if until is not None:
            params += [("until", until.isoformat())]
        if recent_first:
            params += [("reverse", "true")]

        async for bake in self._raw_client.list(
            f"flow/bakes", _parse_bake_payload, params=params
        ):
            yield bake

    async def create_bake(
        self,
        batch: str,
        meta: BakeMeta,
        graphs: Mapping[FullID, Mapping[FullID, AbstractSet[FullID]]],
        params: Optional[Mapping[str, str]] = None,
        name: Optional[str] = None,
        tags: Sequence[str] = (),
    ) -> Bake:
        project_id, _ = await self.get_id()
        graph = {}
        for k1, v1 in graphs.items():
            subgraph = {}
            for k2, v2 in v1.items():
                subgraph[_id_to_json(k2)] = [_id_to_json(it) for it in v2]
            graph[_id_to_json(k1)] = subgraph

        bake_payload = {
            "project_id": project_id,
            "batch": batch,
            "graphs": graph,
            "params": params,
            "name": name,
            "tags": tags,
            "meta": {
                "git_info": (
                    {
                        "sha": meta.git_info.sha,
                        "branch": meta.git_info.branch,
                        "tags": meta.git_info.tags,
                    }
                    if meta.git_info
                    else None
                ),
            },
        }
        return await self._raw_client.create(
            "flow/bakes", bake_payload, _parse_bake_payload
        )

    async def create_cache_entry(
        self,
        task_id: FullID,
        batch: str,
        key: str,
        outputs: Mapping[str, str],
        state: Mapping[str, str],
        raw_id: str,
    ) -> CacheEntry:
        project_id, _ = await self.get_id()
        cache_entry_payload = {
            "project_id": project_id,
            "batch": batch,
            "task_id": _id_to_json(task_id),
            "key": key,
            "outputs": outputs,
            "state": state,
            "raw_id": raw_id,
        }
        return await self._raw_client.create(
            "flow/cache_entries", cache_entry_payload, _parse_cache_entry_payload
        )

    async def delete_cache_entries(
        self,
        batch: Optional[str] = None,
        task_id: Optional[FullID] = None,
    ) -> None:
        project_id, _ = await self.get_id()
        params = [("project_id", project_id)]
        if batch is not None:
            params += [("batch", batch)]
        if task_id is not None:
            params += [("task_id", _id_to_json(task_id))]
        await self._raw_client.delete("flow/cache_entries", params)

    async def list_live_jobs(self) -> AsyncIterator[LiveJob]:
        project_id, _ = await self.get_id()

        async for live_job in self._raw_client.list(
            f"flow/live_jobs",
            _parse_live_job_payload,
            params={"project_id": project_id},
        ):
            yield live_job

    async def create_live_job(
        self,
        yaml_id: str,
        multi: bool,
        tags: Iterable[str],
        raw_id: Optional[str] = None,
    ) -> LiveJob:
        project_id, _ = await self.get_id()
        payload = {
            "project_id": project_id,
            "yaml_id": yaml_id,
            "tags": list(tags),
            "raw_id": raw_id,
            "multi": multi,
        }
        return await self._raw_client.create(
            "flow/live_jobs", payload, _parse_live_job_payload
        )

    async def replace_live_job(
        self,
        yaml_id: str,
        multi: bool,
        tags: Iterable[str],
        raw_id: Optional[str] = None,
    ) -> LiveJob:
        project_id, _ = await self.get_id()
        payload = {
            "project_id": project_id,
            "yaml_id": yaml_id,
            "tags": list(tags),
            "multi": multi,
        }
        if raw_id:
            payload["raw_id"] = raw_id
        return await self._raw_client.update(
            "flow/live_jobs/replace", payload, _parse_live_job_payload
        )

    def live_job(
        self, *, id: Optional[str] = None, yaml_id: Optional[str] = None
    ) -> "LiveJobStorage":
        if id:
            return ApiLiveJobStorage(self._raw_client, id=id)
        else:
            assert yaml_id
            return ApiLiveJobStorage(
                self._raw_client, args=dict(yaml_id=yaml_id, project_storage=self)
            )

    def bake(
        self, *, id: Optional[str] = None, name: Optional[str] = None
    ) -> "BakeStorage":
        if id:
            return ApiBakeStorage(self._raw_client, id=id)
        else:
            assert name
            return ApiBakeStorage(
                self._raw_client, args=dict(name=name, project_storage=self)
            )

    def cache_entry(
        self,
        *,
        id: Optional[str] = None,
        task_id: Optional[FullID] = None,
        batch: Optional[str] = None,
        key: Optional[str] = None,
    ) -> "CacheEntryStorage":
        if id:
            return ApiCacheEntryStorage(self._raw_client, id=id)
        else:
            assert task_id
            assert batch
            assert key
            return ApiCacheEntryStorage(
                self._raw_client,
                args=dict(task_id=task_id, batch=batch, key=key, project_storage=self),
            )


class BakeInitArgs(TypedDict):
    project_storage: ApiProjectStorage
    name: str


class ApiBakeStorage(DeferredIdMixin[Bake, BakeInitArgs], BakeStorage):
    @overload
    def __init__(self, raw_client: RawApiClient, *, id: str):
        pass

    @overload
    def __init__(
        self,
        raw_client: RawApiClient,
        *,
        args: BakeInitArgs,
    ):
        pass

    def __init__(
        self,
        raw_client: RawApiClient,
        *,
        id: Optional[str] = None,
        args: Optional[BakeInitArgs] = None,
    ):
        self._raw_client = raw_client
        self._id = id
        self._args = args

    async def _retrieve_id(self, args: BakeInitArgs) -> Tuple[str, Bake]:
        project_id, _ = await args["project_storage"].get_id()
        bake = await self._raw_client.get(
            "flow/bakes/by_name",
            _parse_bake_payload,
            params={
                "name": args["name"],
                "project_id": project_id,
                "fetch_last_attempt": "1",
            },
        )
        return bake.id, bake

    async def get(self) -> Bake:
        bake_id, bake = await self.get_id()
        if bake:
            return bake
        return await self._raw_client.get(
            f"flow/bakes/{bake_id}",
            _parse_bake_payload,
            params={"fetch_last_attempt": "1"},
        )

    async def list_attempts(self) -> AsyncIterator[Attempt]:
        bake_id, _ = await self.get_id()

        async for attempt in self._raw_client.list(
            f"flow/attempts", _parse_attempt_payload, params={"bake_id": bake_id}
        ):
            yield attempt

    async def create_attempt(
        self,
        configs_meta: ConfigsMeta,
        number: Optional[int] = None,
        executor_id: Optional[str] = None,
        result: TaskStatus = TaskStatus.PENDING,
    ) -> Attempt:
        bake_id, _ = await self.get_id()
        payload = {
            "bake_id": bake_id,
            "number": number,
            "executor_id": executor_id,
            "result": result.value,
            "configs_meta": {
                "workspace": configs_meta.workspace,
                "flow_config_id": configs_meta.flow_config_id,
                "project_config_id": configs_meta.project_config_id,
                "action_config_ids": configs_meta.action_config_ids,
            },
        }
        return await self._raw_client.create(
            f"flow/attempts", payload, _parse_attempt_payload
        )

    async def create_config_file(self, filename: str, content: str) -> ConfigFile:
        bake_id, _ = await self.get_id()
        payload = {
            "bake_id": bake_id,
            "filename": filename,
            "content": content,
        }
        return await self._raw_client.create(
            f"flow/config_files", payload, _parse_config_file_payload
        )

    async def list_bake_images(self) -> AsyncIterator[BakeImage]:
        bake_id, _ = await self.get_id()

        async for attempt in self._raw_client.list(
            f"flow/bake_images", _parse_bake_image_payload, params={"bake_id": bake_id}
        ):
            yield attempt

    async def create_bake_image(
        self,
        yaml_defs: Sequence[FullID],
        ref: str,
        status: ImageStatus = ImageStatus.PENDING,
        context_on_storage: Optional[URL] = None,
        dockerfile_rel: Optional[str] = None,
        builder_job_id: Optional[str] = None,
    ) -> BakeImage:
        bake_id, _ = await self.get_id()
        payload = {
            "bake_id": bake_id,
            "yaml_defs": [_id_to_json(yaml_id) for yaml_id in yaml_defs],
            "ref": ref,
            "status": status.value,
            "context_on_storage": (
                str(context_on_storage) if context_on_storage else None
            ),
            "dockerfile_rel": dockerfile_rel,
            "builder_job_id": builder_job_id,
        }
        return await self._raw_client.create(
            f"flow/bake_images", payload, _parse_bake_image_payload
        )

    def attempt(
        self, *, id: Optional[str] = None, number: Optional[int] = None
    ) -> "AttemptStorage":
        if id:
            return ApiAttemptStorage(self._raw_client, id=id)
        else:
            assert number
            return ApiAttemptStorage(
                self._raw_client, args=dict(bake_storage=self, number=number)
            )

    def config_file(self, *, id: str) -> "ConfigFileStorage":
        return ApiConfigFileStorage(self._raw_client, id=id)

    def bake_image(
        self, *, id: Optional[str] = None, ref: Optional[str] = None
    ) -> "BakeImageStorage":
        if id:
            return ApiBakeImageStorage(self._raw_client, id=id)
        assert ref

        return ApiBakeImageStorage(
            self._raw_client,
            args=dict(
                bake_storage=self,
                ref=ref,
            ),
        )


class CacheEntryInitArgs(TypedDict):
    batch: str
    task_id: FullID
    key: str
    project_storage: ApiProjectStorage


class ApiCacheEntryStorage(
    DeferredIdMixin[CacheEntry, CacheEntryInitArgs], CacheEntryStorage
):
    @overload
    def __init__(self, raw_client: RawApiClient, *, id: str):
        pass

    @overload
    def __init__(
        self,
        raw_client: RawApiClient,
        *,
        args: CacheEntryInitArgs,
    ):
        pass

    def __init__(
        self,
        raw_client: RawApiClient,
        *,
        id: Optional[str] = None,
        args: Optional[CacheEntryInitArgs] = None,
    ):
        self._raw_client = raw_client
        self._id = id
        self._args = args

    async def _retrieve_id(self, args: CacheEntryInitArgs) -> Tuple[str, CacheEntry]:
        project_id, _ = await args["project_storage"].get_id()
        cache_entry = await self._raw_client.get(
            "flow/cache_entries/by_key",
            _parse_cache_entry_payload,
            params={
                "project_id": project_id,
                "batch": args["batch"],
                "task_id": _id_to_json(args["task_id"]),
                "key": args["key"],
            },
        )
        return cache_entry.id, cache_entry

    async def get(self) -> CacheEntry:
        entry_id, entry = await self.get_id()
        if entry:
            return entry
        return await self._raw_client.get(
            f"flow/cache_entries/{entry_id}", _parse_cache_entry_payload
        )


class LiveJobsInitArgs(TypedDict):
    yaml_id: str
    project_storage: ApiProjectStorage


class ApiLiveJobStorage(DeferredIdMixin[LiveJob, LiveJobsInitArgs], LiveJobStorage):
    @overload
    def __init__(self, raw_client: RawApiClient, *, id: str):
        pass

    @overload
    def __init__(
        self,
        raw_client: RawApiClient,
        *,
        args: LiveJobsInitArgs,
    ):
        pass

    def __init__(
        self,
        raw_client: RawApiClient,
        *,
        id: Optional[str] = None,
        args: Optional[LiveJobsInitArgs] = None,
    ):
        self._raw_client = raw_client
        self._id = id
        self._args = args

    async def _retrieve_id(self, args: LiveJobsInitArgs) -> Tuple[str, LiveJob]:
        project_id, _ = await args["project_storage"].get_id()
        live_job = await self._raw_client.get(
            "flow/live_jobs/by_yaml_id",
            _parse_live_job_payload,
            params={
                "project_id": project_id,
                "yaml_id": args["yaml_id"],
            },
        )
        return live_job.id, live_job

    async def get(self) -> LiveJob:
        live_job_id, live_job = await self.get_id()
        if live_job:
            return live_job
        return await self._raw_client.get(
            f"flow/live_jobs/{live_job_id}", _parse_live_job_payload
        )


class AttemptInitArgs(TypedDict):
    bake_storage: ApiBakeStorage
    number: int


class ApiAttemptStorage(DeferredIdMixin[Attempt, AttemptInitArgs], AttemptStorage):
    @overload
    def __init__(self, raw_client: RawApiClient, *, id: str):
        pass

    @overload
    def __init__(
        self,
        raw_client: RawApiClient,
        *,
        args: AttemptInitArgs,
    ):
        pass

    def __init__(
        self,
        raw_client: RawApiClient,
        *,
        id: Optional[str] = None,
        args: Optional[AttemptInitArgs] = None,
    ):
        self._raw_client = raw_client
        self._id = id
        self._args = args

    async def _retrieve_id(self, args: AttemptInitArgs) -> Tuple[str, Attempt]:
        bake_id, _ = await args["bake_storage"].get_id()
        attempt = await self._raw_client.get(
            "flow/attempts/by_number",
            _parse_attempt_payload,
            params={
                "bake_id": bake_id,
                "number": str(args["number"]),
            },
        )
        return attempt.id, attempt

    async def get(self) -> Attempt:
        attempt_id, attempt = await self.get_id()
        if attempt:
            return attempt
        return await self._raw_client.get(
            f"flow/attempts/{attempt_id}", _parse_attempt_payload
        )

    async def update(
        self,
        *,
        executor_id: Union[Optional[str], Type[_Unset]] = _Unset,
        result: Union[TaskStatus, Type[_Unset]] = _Unset,
    ) -> Attempt:
        # TODO: Use PATCH when implemented on server
        attempt = await self.get()
        if executor_id is not _Unset:
            attempt = replace(
                attempt, executor_id=executor_id  # type: ignore[arg-type]
            )
        if result is not _Unset:
            attempt = replace(attempt, result=result)  # type: ignore[arg-type]
        payload = {
            "bake_id": attempt.bake_id,
            "number": attempt.number,
            "executor_id": attempt.executor_id,
            "result": attempt.result.value,
            "configs_meta": {
                "workspace": attempt.configs_meta.workspace,
                "flow_config_id": attempt.configs_meta.flow_config_id,
                "project_config_id": attempt.configs_meta.project_config_id,
                "action_config_ids": attempt.configs_meta.action_config_ids,
            },
        }
        return await self._raw_client.update(
            "flow/attempts/replace", data=payload, mapper=_parse_attempt_payload
        )

    async def list_tasks(self) -> AsyncIterator[Task]:
        attempt_id, _ = await self.get_id()

        async for task in self._raw_client.list(
            f"flow/tasks", _parse_task_payload, params={"attempt_id": attempt_id}
        ):
            yield task

    async def create_task(
        self,
        yaml_id: FullID,
        raw_id: Optional[str],
        status: Union[TaskStatusItem, TaskStatus],
        outputs: Optional[Mapping[str, str]] = None,
        state: Optional[Mapping[str, str]] = None,
    ) -> Task:
        if isinstance(status, TaskStatus):
            status = TaskStatusItem(when=_now(), status=status)
        attempt_id, _ = await self.get_id()
        task_payload = {
            "attempt_id": attempt_id,
            "yaml_id": _id_to_json(yaml_id),
            "raw_id": raw_id,
            "outputs": outputs,
            "state": state,
            "statuses": [
                {
                    "created_at": _dt2str(status.when),
                    "status": status.status.value,
                }
            ],
        }
        return await self._raw_client.create(
            "flow/tasks", task_payload, _parse_task_payload
        )

    def task(
        self, *, id: Optional[str] = None, yaml_id: Optional[FullID] = None
    ) -> "TaskStorage":
        if id:
            return ApiTaskStorage(self._raw_client, id=id)
        else:
            assert yaml_id

            return ApiTaskStorage(
                self._raw_client,
                args=dict(
                    yaml_id=yaml_id,
                    attempt_storage=self,
                ),
            )


class ApiConfigFileStorage(ConfigFileStorage):
    def __init__(
        self,
        raw_client: RawApiClient,
        id: str,
    ):
        self._raw_client = raw_client
        self._id = id

    async def get(self) -> ConfigFile:
        return await self._raw_client.get(
            f"flow/config_files/{self._id}", _parse_config_file_payload
        )


class BakeImageInitArgs(TypedDict):
    ref: str
    bake_storage: ApiBakeStorage


class ApiBakeImageStorage(
    DeferredIdMixin[BakeImage, BakeImageInitArgs], BakeImageStorage
):
    @overload
    def __init__(self, raw_client: RawApiClient, *, id: str):
        pass

    @overload
    def __init__(
        self,
        raw_client: RawApiClient,
        *,
        args: BakeImageInitArgs,
    ):
        pass

    def __init__(
        self,
        raw_client: RawApiClient,
        *,
        id: Optional[str] = None,
        args: Optional[BakeImageInitArgs] = None,
    ):
        self._raw_client = raw_client
        self._id = id
        self._args = args

    async def _retrieve_id(self, args: BakeImageInitArgs) -> Tuple[str, BakeImage]:
        bake_id, _ = await args["bake_storage"].get_id()
        bake_image = await self._raw_client.get(
            "flow/bake_images/by_ref",
            _parse_bake_image_payload,
            params={
                "bake_id": bake_id,
                "ref": args["ref"],
            },
        )
        return bake_image.id, bake_image

    async def get(self) -> BakeImage:
        bake_image_id, bake_image = await self.get_id()
        if bake_image:
            return bake_image
        return await self._raw_client.get(
            f"flow/bake_images/{bake_image_id}", _parse_bake_image_payload
        )

    async def update(
        self,
        *,
        status: Union[ImageStatus, Type[_Unset]] = _Unset,
        builder_job_id: Union[Optional[str], Type[_Unset]] = _Unset,
    ) -> BakeImage:
        bake_image_id, _ = await self.get_id()

        patch_data: Dict[str, Any] = {}
        if status != _Unset:
            patch_data["status"] = status
        if builder_job_id != _Unset:
            patch_data["builder_job_id"] = builder_job_id
        return await self._raw_client.patch(
            f"flow/bake_images/{bake_image_id}", patch_data, _parse_bake_image_payload
        )


class TaskInitArgs(TypedDict):
    attempt_storage: ApiAttemptStorage
    yaml_id: FullID


class ApiTaskStorage(DeferredIdMixin[Task, TaskInitArgs], TaskStorage):
    @overload
    def __init__(self, raw_client: RawApiClient, *, id: str):
        pass

    @overload
    def __init__(
        self,
        raw_client: RawApiClient,
        *,
        args: TaskInitArgs,
    ):
        pass

    def __init__(
        self,
        raw_client: RawApiClient,
        *,
        id: Optional[str] = None,
        args: Optional[TaskInitArgs] = None,
    ):
        self._raw_client = raw_client
        self._id = id
        self._args = args

    async def _retrieve_id(self, args: TaskInitArgs) -> Tuple[str, Task]:
        attempt_id, _ = await args["attempt_storage"].get_id()
        task = await self._raw_client.get(
            "flow/tasks/by_yaml_id",
            _parse_task_payload,
            params={
                "attempt_id": attempt_id,
                "yaml_id": ".".join(args["yaml_id"]),
            },
        )
        return task.id, task

    async def get(self) -> Task:
        task_id, task = await self.get_id()
        if task:
            return task
        return await self._raw_client.get(f"flow/tasks/{task_id}", _parse_task_payload)

    async def update(
        self,
        *,
        outputs: Union[Optional[Mapping[str, str]], Type[_Unset]] = _Unset,
        state: Union[Optional[Mapping[str, str]], Type[_Unset]] = _Unset,
        new_status: Optional[Union[TaskStatusItem, TaskStatus]] = None,
    ) -> Task:
        # TODO: Use PATCH when implemented on server
        task = await self.get()
        if outputs is not _Unset:
            task = replace(task, outputs=outputs)  # type: ignore[arg-type]
        if state is not _Unset:
            task = replace(task, state=state)  # type: ignore[arg-type]
        if new_status:
            if isinstance(new_status, TaskStatus):
                new_status = TaskStatusItem(when=_now(), status=new_status)
            task = replace(task, statuses=[*task.statuses, new_status])
        task_payload = {
            "attempt_id": task.attempt_id,
            "yaml_id": _id_to_json(task.yaml_id),
            "raw_id": task.raw_id,
            "outputs": task.outputs,
            "state": task.state,
            "statuses": [
                {
                    "created_at": _dt2str(item.when),
                    "status": item.status.value,
                }
                for item in task.statuses
            ],
        }
        return await self._raw_client.update(
            f"flow/tasks/replace", task_payload, _parse_task_payload
        )



import apolo_sdk
import click
from contextlib import AsyncExitStack
from typing import List, Optional, Sequence, Tuple

from apolo_flow.cli.click_types import (
    LIVE_JOB,
    LIVE_JOB_OR_ALL,
    PROJECT,
    LiveJobSuffixType,
)
from apolo_flow.cli.utils import argument, option, wrap_async
from apolo_flow.live_runner import LiveRunner

from ..storage.api import ApiStorage
from ..storage.base import Storage
from .root import Root


@click.command()
@wrap_async()
async def ps(
    root: Root,
) -> None:
    """List all jobs"""
    async with AsyncExitStack() as stack:
        client = await stack.enter_async_context(apolo_sdk.get())
        storage: Storage = await stack.enter_async_context(ApiStorage(client))
        runner = await stack.enter_async_context(
            LiveRunner(root.config_dir, root.console, client, storage, root)
        )
        await runner.ps()


@click.command()
@option("-s", "--suffix", help="Optional suffix for multi-jobs")
@option(
    "-p",
    "--param",
    type=(str, str),
    multiple=True,
    help="Set params of the batch config",
)
@option(
    "--dry-run",
    is_flag=True,
    default=False,
    help="Print run command instead of starting job.",
)
@argument("job-id", type=LIVE_JOB)
@argument("args", nargs=-1)
@wrap_async()
async def run(
    root: Root,
    job_id: str,
    suffix: Optional[str],
    dry_run: bool,
    args: Optional[Tuple[str]],
    param: List[Tuple[str, str]],
) -> None:
    """Run a job.

    RUN job JOB-ID or ATTACH to it if the job is already running

    For multi-jobs an explicit job suffix can be used with explicit job arguments.
    """
    if args:
        root.console.print(
            "[yellow]args are deprecated, use --param instead",
        )
    async with AsyncExitStack() as stack:
        client = await stack.enter_async_context(apolo_sdk.get())
        storage: Storage = await stack.enter_async_context(ApiStorage(client))
        runner = await stack.enter_async_context(
            LiveRunner(root.config_dir, root.console, client, storage, root, dry_run)
        )
        await runner.run(
            job_id,
            suffix=suffix,
            args=args,
            params={key: value for key, value in param},
        )


@click.command()
@argument("job-id", type=LIVE_JOB)
@argument("suffix", required=False, type=LiveJobSuffixType(job_id_param_name="job_id"))
@wrap_async()
async def logs(
    root: Root,
    job_id: str,
    suffix: Optional[str],
) -> None:
    """Print logs.

    Display logs for JOB-ID
    """
    async with AsyncExitStack() as stack:
        client = await stack.enter_async_context(apolo_sdk.get())
        storage: Storage = await stack.enter_async_context(ApiStorage(client))
        runner = await stack.enter_async_context(
            LiveRunner(root.config_dir, root.console, client, storage, root)
        )
        await runner.logs(job_id, suffix)


@click.command()
@argument("job-id", type=LIVE_JOB)
@argument("suffix", required=False, type=LiveJobSuffixType(job_id_param_name="job_id"))
@wrap_async()
async def status(
    root: Root,
    job_id: str,
    suffix: Optional[str],
) -> None:
    """Show job status.

    Print status for JOB-ID
    """
    async with AsyncExitStack() as stack:
        client = await stack.enter_async_context(apolo_sdk.get())
        storage: Storage = await stack.enter_async_context(ApiStorage(client))
        runner = await stack.enter_async_context(
            LiveRunner(root.config_dir, root.console, client, storage, root)
        )
        await runner.status(job_id, suffix)


@click.command()
@argument("job-id", type=LIVE_JOB_OR_ALL)
@argument("suffix", required=False, type=LiveJobSuffixType(job_id_param_name="job_id"))
@wrap_async()
async def kill(
    root: Root,
    job_id: str,
    suffix: Optional[str],
) -> None:
    """Kill a job.

    Kill JOB-ID, use `kill ALL` for killing all jobs."""
    async with AsyncExitStack() as stack:
        client = await stack.enter_async_context(apolo_sdk.get())
        storage: Storage = await stack.enter_async_context(ApiStorage(client))
        runner = await stack.enter_async_context(
            LiveRunner(root.config_dir, root.console, client, storage, root)
        )
        if job_id != "ALL":
            await runner.kill(job_id, suffix)
        else:
            if suffix is not None:
                raise click.BadArgumentUsage(
                    "Suffix is not supported when killing ALL jobs"
                )
            await runner.kill_all()


@click.command()
@argument("flow_ids", type=PROJECT, nargs=-1, required=True)
@wrap_async()
async def delete_flow(
    root: Root,
    flow_ids: Sequence[str],
) -> None:
    """Completely remove flow with all related entities"""
    async with AsyncExitStack() as stack:
        client = await stack.enter_async_context(apolo_sdk.get())
        storage: Storage = await stack.enter_async_context(ApiStorage(client))
        for flow_id in flow_ids:
            await storage.project(yaml_id=flow_id).delete()
            if root.verbosity >= 0:
                root.console.print(f"Flow '{flow_id}' was successfully removed.")



import asyncio
import logging
import os
import shlex
import sys
import textwrap
from apolo_sdk import (
    Client,
    DiskVolume,
    EnvParseResult,
    HTTPPort,
    JobDescription,
    JobRestartPolicy,
    JobStatus,
    Preset,
    RemoteImage,
    ResourceNotFound,
    SecretFile,
    Tag,
    Volume,
    VolumeParseResult,
)
from collections import defaultdict
from contextlib import asynccontextmanager
from datetime import datetime, timedelta, timezone
from rich.console import Console
from rich.panel import Panel
from rich.progress import BarColumn, Progress, TaskID, TextColumn
from typing import (
    AbstractSet,
    AsyncIterator,
    Awaitable,
    Callable,
    Dict,
    Generic,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)
from yarl import URL

from . import ast
from .colored_topo_sorter import ColoredTopoSorter
from .commands import CmdProcessor
from .config_loader import BatchRemoteCL
from .context import (
    BaseBatchContext,
    DepCtx,
    EarlyImageCtx,
    ImageCtx,
    LocallyPreparedInfo,
    LocalTask,
    NeedsCtx,
    RunningBatchActionFlow,
    RunningBatchBase,
    RunningBatchFlow,
    StateCtx,
    Task,
    TaskMeta,
)
from .expr import EvalError
from .storage.base import (
    Attempt,
    AttemptStorage,
    Bake,
    BakeImage,
    BakeImageStorage,
    BakeStorage,
    ConfigsMeta,
    ProjectStorage,
    Storage,
    Task as StorageTask,
    TaskStatusItem,
    _Unset,
)
from .types import AlwaysT, FullID, ImageStatus, RemotePath, TaskStatus
from .utils import (
    TERMINATED_JOB_STATUSES,
    RetryConfig,
    fmt_id,
    fmt_raw_id,
    fmt_status,
    retry,
)


log = logging.getLogger(__name__)


class NotFinished(ValueError):
    pass


async def get_running_flow(
    bake: Bake, client: Client, storage: BakeStorage, configs_meta: ConfigsMeta
) -> RunningBatchFlow:
    config_loader = BatchRemoteCL(
        configs_meta,
        client=client,
        bake_storage=storage,
    )
    local_info = LocallyPreparedInfo(
        children_info={},
        early_images={},
        git_info=bake.meta.git_info,
    )
    async for image in storage.list_bake_images():
        for yaml_def in image.yaml_defs:
            *prefix, image_id = yaml_def
            sub = local_info
            for part in prefix:
                if part not in sub.children_info:
                    new_sub = LocallyPreparedInfo(
                        children_info={},
                        early_images={},
                        git_info=bake.meta.git_info,
                    )
                    assert isinstance(sub.children_info, dict)
                    sub.children_info[part] = new_sub
                sub = sub.children_info[part]
            assert isinstance(sub.early_images, dict)
            dockerfile = None
            if image.context_on_storage and image.dockerfile_rel:
                dockerfile = image.context_on_storage / str(image.dockerfile_rel)
            sub.early_images[image_id] = EarlyImageCtx(
                id=image_id,
                ref=image.ref,
                context=image.context_on_storage,
                dockerfile=dockerfile,
                dockerfile_rel=(
                    RemotePath(image.dockerfile_rel) if image.dockerfile_rel else None
                ),
            )
    return await RunningBatchFlow.create(
        config_loader=config_loader,
        batch=bake.batch,
        bake_id=bake.id,
        params=bake.params,
        local_info=local_info,
    )


_T = TypeVar("_T")


class Graph(Generic[_T]):
    """TopologicalSorter that allows to embed sub graphs in nodes

    Instance of this graph is constructed by provided simple graph
    specification. After that, additional graphs can be embedded
    to nodes of this (or previously embedded) graph. Each node
    has address of type FullID, where each component is names
    node in some graphs. For example, node id ('top', 'node')
    is a node of graph embedded to node 'top' of the main top
    level graph.

    Node is considered ready when it is ready inside its graph
    and the parent node is ready. Top level graph is ready by
    default.

    Each graph can be annotated with some metadata.
    """

    def __init__(self, graph: Mapping[str, Mapping[str, ast.NeedsLevel]], meta: _T):
        topo = ColoredTopoSorter(graph)
        self._topos: Dict[FullID, ColoredTopoSorter[str, ast.NeedsLevel]] = {(): topo}
        self._sizes: Dict[FullID, int] = {(): len(graph)}
        self._metas: Dict[FullID, _T] = {(): meta}
        self._ready: Set[FullID] = set()
        self._done: Set[FullID] = set()
        self._ready_to_mark_embeds: Dict[ast.NeedsLevel, Set[FullID]] = defaultdict(set)
        self._process_ready(())

    @property
    def is_active(self) -> bool:
        log.debug(f"Graph: is_active check, self._ready={self._ready}")
        return bool(self._ready)

    def get_ready(self) -> AbstractSet[FullID]:
        return set(self._ready)

    def get_ready_with_meta(self) -> Iterable[Tuple[FullID, _T]]:
        for ready in set(self._ready):
            yield ready, self._metas[ready[:-1]]

    def get_graph_data(self, node: FullID) -> _T:
        return self._metas[node]

    def get_meta(self, node: FullID) -> _T:
        prefix = node[:-1]
        assert prefix in self._topos, f"Parent graph for node {node} not found"
        return self._metas[prefix]

    def get_ready_to_mark_running_embeds(self) -> AbstractSet[FullID]:
        return set(self._ready_to_mark_embeds[ast.NeedsLevel.RUNNING])

    def get_ready_to_mark_done_embeds(self) -> AbstractSet[FullID]:
        return set(self._ready_to_mark_embeds[ast.NeedsLevel.COMPLETED])

    def get_size(self, prefix: FullID) -> int:
        return self._sizes[prefix]

    def _mark_node(self, node: FullID, level: ast.NeedsLevel) -> None:
        log.debug(f"Graph: marking node {node} with {level}")

        prefix, item = node[:-1], node[-1]
        assert prefix in self._topos, f"Parent graph for node {node} not found"
        topo = self._topos[prefix]
        if topo.is_colored(item, level):
            return

        if node in self._topos:
            assert node in self._ready_to_mark_embeds[level], (
                f"Node '{node}' cannot be marked to {level} when "
                f"embeded graph is not all marked with {level}"
            )
            self._ready_to_mark_embeds[level].remove(node)

        topo.mark(item, level)
        if topo.is_all_colored(level) and prefix != ():
            self._ready_to_mark_embeds[level].add(prefix)
        else:
            self._process_ready(prefix)

    def mark_running(self, node: FullID) -> None:
        self._mark_node(node, ast.NeedsLevel.RUNNING)

    def mark_done(self, node: FullID) -> None:
        self._mark_node(node, ast.NeedsLevel.COMPLETED)
        self._ready.remove(node)
        self._done.add(node)

    def embed(
        self, node: FullID, graph: Mapping[str, Mapping[str, ast.NeedsLevel]], meta: _T
    ) -> None:
        log.debug(f"Graph: embedding to node {node} graph {graph}")
        prefix = node[:-1]
        assert node not in self._done, f"Cannot embed to already done node {node}"
        assert prefix in self._topos, f"Parent graph for node {node} not found"
        self._topos[node] = ColoredTopoSorter(graph)
        self._sizes[node] = len(graph)
        self._metas[node] = meta
        if node in self._ready:
            self._process_ready(node)

    def _process_ready(self, prefix: FullID) -> None:
        # Collects ready node from subgraphs recursively
        for ready in self._topos[prefix].get_ready():
            node = prefix + (ready,)
            self._ready.add(node)
            if node in self._topos:
                self._process_ready(node)


class BakeTasksManager:
    def __init__(self) -> None:
        self._tasks: Dict[FullID, StorageTask] = {}

    @property
    def tasks(self) -> Mapping[FullID, StorageTask]:
        return self._tasks

    def list_unfinished_raw_tasks(self) -> Iterable[Tuple[StorageTask, str]]:
        for task in self._tasks.values():
            if not task.status.is_finished and task.raw_id:
                yield task, task.raw_id

    def update(self, task: StorageTask) -> None:
        self._tasks[task.yaml_id] = task

    def count_unfinished(self, prefix: FullID = ()) -> int:
        def _with_prefix(full_id: FullID) -> bool:
            return all(x == y for x, y in zip(prefix, full_id))

        return sum(
            1
            for v in self._tasks.values()
            if v.raw_id and _with_prefix(v.yaml_id) and not v.status.is_finished
        )

    def build_needs(self, prefix: FullID, deps: AbstractSet[str]) -> NeedsCtx:
        needs = {}
        for dep_id in deps:
            full_id = prefix + (dep_id,)
            dep = self._tasks.get(full_id)
            if dep is None or not dep.status.is_finished:
                raise NotFinished(full_id)
            assert dep.outputs is not None
            status = TaskStatus(dep.status)
            if status == TaskStatus.CACHED:
                status = TaskStatus.SUCCEEDED
            needs[dep_id] = DepCtx(status, dep.outputs)
        return needs

    def build_state(self, prefix: FullID, state_from: Optional[str]) -> StateCtx:
        if state_from is None:
            return {}
        full_id = prefix + (state_from,)
        dep = self._tasks.get(full_id)
        if dep is None or not dep.status.is_finished:
            raise NotFinished(full_id)
        assert dep.state is not None
        return dep.state


class ImageBuildCommandError(Exception):
    def __init__(self, exit_code: int, stdout: str, stderr: str) -> None:
        self.exit_code = exit_code
        self.stdout = stdout
        self.stderr = stderr

    def __str__(self) -> str:
        return (
            f"Image builder command failed:\n"
            f"exit code: {self.exit_code}\n"
            f"stdout: \n{textwrap.indent(self.stdout, '  ')}\n"
            f"stderr: \n{textwrap.indent(self.stderr, '  ')}"
        )


async def start_image_build(*cmd: str) -> str:
    log.debug(f"start_image_build: executing {cmd}")
    subprocess = await asyncio.create_subprocess_exec(
        *cmd,
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    assert subprocess.stdout
    assert subprocess.stderr
    stdout = ""
    while subprocess.returncode is None:
        log.debug(f"start_image_build: reading line...")
        line = (await subprocess.stdout.readline()).decode()
        log.debug(f"start_image_build: read line: {line}")
        stdout += line
        if line.startswith("Job ID: "):
            # Job is started, we can freely kill apolo-cli
            subprocess.terminate()
            return line[len("Job ID: ") :].strip()

    stdout += (await subprocess.stdout.read()).decode()
    stderr = (await subprocess.stderr.read()).decode()
    raise ImageBuildCommandError(
        exit_code=subprocess.returncode,
        stdout=stdout,
        stderr=stderr,
    )


class RetryReadNeuroClient(RetryConfig):
    def __init__(self, client: Client) -> None:
        super().__init__()
        self._client = client

    async def job_start(
        self,
        *,
        image: RemoteImage,
        preset_name: str,
        entrypoint: Optional[str] = None,
        command: Optional[str] = None,
        working_dir: Optional[str] = None,
        http: Optional[HTTPPort] = None,
        env: Optional[Mapping[str, str]] = None,
        volumes: Sequence[Volume] = (),
        secret_env: Optional[Mapping[str, URL]] = None,
        secret_files: Sequence[SecretFile] = (),
        disk_volumes: Sequence[DiskVolume] = (),
        tty: bool = False,
        shm: bool = False,
        name: Optional[str] = None,
        tags: Sequence[str] = (),
        description: Optional[str] = None,
        pass_config: bool = False,
        wait_for_jobs_quota: bool = False,
        schedule_timeout: Optional[float] = None,
        restart_policy: JobRestartPolicy = JobRestartPolicy.NEVER,
        life_span: Optional[float] = None,
        privileged: bool = False,
        project_name: Optional[str] = None,
    ) -> JobDescription:
        return await self._client.jobs.start(
            image=image,
            preset_name=preset_name,
            entrypoint=entrypoint,
            command=command,
            working_dir=working_dir,
            http=http,
            env=env,
            volumes=volumes,
            secret_env=secret_env,
            secret_files=secret_files,
            disk_volumes=disk_volumes,
            tty=tty,
            shm=shm,
            name=name,
            tags=tags,
            description=description,
            pass_config=pass_config,
            wait_for_jobs_quota=wait_for_jobs_quota,
            schedule_timeout=schedule_timeout,
            restart_policy=restart_policy,
            life_span=life_span,
            privileged=privileged,
            project_name=project_name,
        )

    @retry
    async def job_status(self, raw_id: str) -> JobDescription:
        return await self._client.jobs.status(raw_id)

    @retry
    async def _job_logs(self, raw_id: str) -> List[bytes]:
        chunks = []
        async for chunk in self._client.jobs.monitor(raw_id):
            chunks.append(chunk)
        return chunks

    async def job_logs(self, raw_id: str) -> AsyncIterator[bytes]:
        for chunk in await self._job_logs(raw_id):
            yield chunk

    async def job_kill(self, raw_id: str) -> None:
        await self._client.jobs.kill(raw_id)

    @retry
    async def image_tag_info(self, remote_image: RemoteImage) -> Tag:
        return await self._client.images.tag_info(remote_image)

    def parse_remote_image(self, image: str) -> RemoteImage:
        return self._client.parse.remote_image(image)

    @property
    def config_presets(self) -> Mapping[str, Preset]:
        return self._client.config.presets

    def parse_envs(
        self, env: Sequence[str], env_file: Sequence[str] = ()
    ) -> EnvParseResult:
        return self._client.parse.envs(env, env_file)

    def parse_volumes(self, volume: Sequence[str]) -> VolumeParseResult:
        return self._client.parse.volumes(volume)


class BatchExecutor:
    transient_progress: bool = False
    run_builder_job: Callable[..., Awaitable[str]] = staticmethod(  # type: ignore
        start_image_build
    )

    def __init__(
        self,
        console: Console,
        flow: RunningBatchFlow,
        bake: Bake,
        attempt: Attempt,
        client: RetryReadNeuroClient,
        storage: AttemptStorage,
        bake_storage: BakeStorage,
        project_storage: ProjectStorage,
        *,
        polling_timeout: Optional[float] = 1,
        project_role: Optional[str] = None,
    ) -> None:
        self._progress = Progress(
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TextColumn("[progress.percentage]{task.percentage:>3.0f}%"),
            TextColumn("[progress.remaining]{task.elapsed:.0f} sec"),
            console=console,
            auto_refresh=False,
            transient=self.transient_progress,
            redirect_stderr=True,
        )
        self._top_flow = flow
        self._bake = bake
        self._attempt = attempt
        self._client = client
        self._storage = storage
        self._project_storage = project_storage
        self._bake_storage = bake_storage
        self._graphs: Graph[RunningBatchBase[BaseBatchContext]] = Graph(
            flow.graph, flow
        )
        self._tasks_mgr = BakeTasksManager()
        self._is_cancelling = False
        self._project_role = project_role

        self._run_builder_job = self.run_builder_job

        # A note about default value:
        # AS: I have no idea what timeout is better;
        # too short value bombards servers,
        # too long timeout makes the waiting longer than expected
        # The missing events subsystem would be great for this task :)
        self._polling_timeout = polling_timeout
        self._bars: Dict[FullID, TaskID] = {}

    @classmethod
    @asynccontextmanager
    async def create(
        cls,
        console: Console,
        bake_id: str,
        client: Client,
        storage: Storage,
        *,
        polling_timeout: Optional[float] = 1,
        project_role: Optional[str] = None,
    ) -> AsyncIterator["BatchExecutor"]:
        storage = storage.with_retry_read()

        console.log("Fetch bake data")
        bake_storage = storage.bake(id=bake_id)
        bake = await bake_storage.get()
        if bake.last_attempt:
            attempt = bake.last_attempt
            attempt_storage = bake_storage.attempt(id=bake.last_attempt.id)
        else:
            attempt_storage = bake_storage.last_attempt()
            attempt = await attempt_storage.get()

        console.log("Fetch configs metadata")
        flow = await get_running_flow(bake, client, bake_storage, attempt.configs_meta)

        console.log(f"Execute attempt #{attempt.number}")

        ret = cls(
            console=console,
            flow=flow,
            bake=bake,
            attempt=attempt,
            client=RetryReadNeuroClient(client),
            storage=attempt_storage,
            bake_storage=bake_storage,
            project_storage=storage.project(id=bake.project_id),
            polling_timeout=polling_timeout,
            project_role=project_role,
        )
        ret._start()
        try:
            yield ret
        finally:
            ret._stop()

    def _start(self) -> None:
        pass

    def _stop(self) -> None:
        pass

    async def _refresh_attempt(self) -> None:
        self._attempt = await self._storage.get()

    def _only_completed_needs(
        self, needs: Mapping[str, ast.NeedsLevel]
    ) -> AbstractSet[str]:
        return {
            task_id
            for task_id, level in needs.items()
            if level == ast.NeedsLevel.COMPLETED
        }

    # Exact task/action contexts helpers
    async def _get_meta(self, full_id: FullID) -> TaskMeta:
        prefix, tid = full_id[:-1], full_id[-1]
        flow = self._graphs.get_meta(full_id)
        needs = self._tasks_mgr.build_needs(
            prefix, self._only_completed_needs(flow.graph[tid])
        )
        state = self._tasks_mgr.build_state(prefix, await flow.state_from(tid))
        return await flow.get_meta(tid, needs=needs, state=state)

    async def _get_task(self, full_id: FullID) -> Task:
        prefix, tid = full_id[:-1], full_id[-1]
        flow = self._graphs.get_meta(full_id)
        needs = self._tasks_mgr.build_needs(
            prefix, self._only_completed_needs(flow.graph[tid])
        )
        state = self._tasks_mgr.build_state(prefix, await flow.state_from(tid))
        return await flow.get_task(prefix, tid, needs=needs, state=state)

    async def _get_local(self, full_id: FullID) -> LocalTask:
        prefix, tid = full_id[:-1], full_id[-1]
        flow = self._graphs.get_meta(full_id)
        needs = self._tasks_mgr.build_needs(
            prefix, self._only_completed_needs(flow.graph[tid])
        )
        return await flow.get_local(tid, needs=needs)

    async def _get_action(self, full_id: FullID) -> RunningBatchActionFlow:
        prefix, tid = full_id[:-1], full_id[-1]
        flow = self._graphs.get_meta(full_id)
        needs = self._tasks_mgr.build_needs(
            prefix, self._only_completed_needs(flow.graph[tid])
        )
        return await flow.get_action(tid, needs=needs)

    async def _get_image(self, bake_image: BakeImage) -> Optional[ImageCtx]:
        actions = []
        full_id_to_image: Dict[FullID, ImageCtx] = {}
        for yaml_def in bake_image.yaml_defs:
            *prefix, image_id = yaml_def
            actions.append(".".join(prefix))
            try:
                flow = self._graphs.get_graph_data(tuple(prefix))
            except KeyError:
                pass
            else:
                full_id_to_image[yaml_def] = flow.images[image_id]
        image_ctx = None
        for _it in full_id_to_image.values():
            image_ctx = _it  # Return any context
            break
        if not all(image_ctx == it for it in full_id_to_image.values()):
            locations_str = "\n".join(
                f" - {'.'.join(full_id)}" for full_id in full_id_to_image.keys()
            )
            self._progress.log(
                f"[red]Warning:[/red] definitions for image with"
                f" ref [b]{bake_image.ref}[/b] differ:\n"
                f"{locations_str}"
            )
        if len(full_id_to_image.values()) == 0:
            actions_str = ", ".join(f"'{it}'" for it in actions)
            self._progress.log(
                f"[red]Warning:[/red] image with ref "
                f"[b]{bake_image.ref}[/b] is referred "
                + (
                    f"before action {actions_str} "
                    "that holds its definition was able to run"
                    if len(actions) == 1
                    else f"before actions {actions_str} "
                    "that hold its definition were able to run"
                )
            )
        return image_ctx

    # Graph helpers

    async def _embed_action(self, full_id: FullID) -> None:
        action = await self._get_action(full_id)
        self._graphs.embed(full_id, action.graph, action)
        task_id = self._progress.add_task(
            ".".join(full_id),
            completed=0,
            total=self._graphs.get_size(full_id) * 3,
        )
        self._bars[full_id] = task_id

    def _advance_progress_bar(
        self, old_task: Optional[StorageTask], task: StorageTask
    ) -> None:
        def _state_to_progress(status: TaskStatus) -> int:
            if status.is_pending:
                return 1
            if status.is_running:
                return 2
            return 3

        if old_task:
            advance = _state_to_progress(task.status) - _state_to_progress(
                old_task.status
            )
        else:
            advance = _state_to_progress(task.status)
        progress_bar_id = self._bars[task.yaml_id[:-1]]
        self._progress.update(progress_bar_id, advance=advance, refresh=True)

    def _update_graph_coloring(self, task: StorageTask) -> None:
        if task.status.is_running or task.status.is_finished:
            self._graphs.mark_running(task.yaml_id)
        if task.status.is_finished:
            self._graphs.mark_done(task.yaml_id)

    async def _log_task_status_change(self, task: StorageTask) -> None:
        flow = self._graphs.get_meta(task.yaml_id)
        in_flow_id = task.yaml_id[-1]
        if await flow.is_action(in_flow_id):
            log_args = ["Action"]
        elif await flow.is_local(in_flow_id):
            log_args = ["Local action"]
        else:
            log_args = ["Task"]
        log_args.append(fmt_id(task.yaml_id))
        if task.raw_id:
            log_args.append(fmt_raw_id(task.raw_id))
        log_args += ["is", fmt_status(task.status)]
        if task.status.is_finished:
            if task.outputs:
                log_args += ["with following outputs:"]
            else:
                log_args += ["(no outputs)"]
        self._progress.log(*log_args)
        if task.outputs:
            for key, value in task.outputs.items():
                self._progress.log(f"  {key}: {value}")

    async def _create_task(
        self,
        yaml_id: FullID,
        raw_id: Optional[str],
        status: Union[TaskStatusItem, TaskStatus],
        outputs: Optional[Mapping[str, str]] = None,
        state: Optional[Mapping[str, str]] = None,
    ) -> StorageTask:
        task = await self._storage.create_task(yaml_id, raw_id, status, outputs, state)
        self._advance_progress_bar(None, task)
        self._update_graph_coloring(task)
        await self._log_task_status_change(task)
        self._tasks_mgr.update(task)
        return task

    async def _update_task(
        self,
        yaml_id: FullID,
        *,
        outputs: Union[Optional[Mapping[str, str]], Type[_Unset]] = _Unset,
        state: Union[Optional[Mapping[str, str]], Type[_Unset]] = _Unset,
        new_status: Optional[Union[TaskStatusItem, TaskStatus]] = None,
    ) -> StorageTask:
        old_task = self._tasks_mgr.tasks[yaml_id]
        task = await self._storage.task(id=old_task.id).update(
            outputs=outputs,
            state=state,
            new_status=new_status,
        )
        self._advance_progress_bar(old_task, task)
        self._update_graph_coloring(task)
        await self._log_task_status_change(task)
        self._tasks_mgr.update(task)
        return task

    # New tasks processers

    async def _skip_task(self, full_id: FullID) -> None:
        log.debug(f"BatchExecutor: skipping task {full_id}")
        await self._create_task(
            yaml_id=full_id,
            status=TaskStatus.SKIPPED,
            raw_id=None,
            outputs={},
            state={},
        )

    async def _process_local(self, full_id: FullID) -> None:
        raise ValueError("Processing of local actions is not supported")

    async def _process_action(self, full_id: FullID) -> None:
        log.debug(f"BatchExecutor: processing action {full_id}")
        await self._create_task(
            yaml_id=full_id,
            status=TaskStatus.PENDING,
            raw_id=None,
        )
        await self._embed_action(full_id)

    async def _process_task(self, full_id: FullID) -> None:
        log.debug(f"BatchExecutor: processing task {full_id}")
        task = await self._get_task(full_id)

        # Check is task fits max_parallel
        for n in range(1, len(full_id) + 1):
            node = full_id[:n]
            prefix = node[:-1]
            if self._tasks_mgr.count_unfinished(prefix) >= task.strategy.max_parallel:
                return

        cache_strategy = task.cache.strategy
        storage_task: Optional[StorageTask] = None
        if cache_strategy == ast.CacheStrategy.DEFAULT:
            try:
                cache_entry = await self._project_storage.cache_entry(
                    task_id=full_id, batch=self._bake.batch, key=task.caching_key
                ).get()
            except ResourceNotFound:
                pass
            else:
                now = datetime.now(timezone.utc)
                eol = cache_entry.created_at + timedelta(task.cache.life_span)
                if now < eol:
                    storage_task = await self._create_task(
                        yaml_id=full_id,
                        raw_id=cache_entry.raw_id,
                        status=TaskStatus.CACHED,
                        outputs=cache_entry.outputs,
                        state=cache_entry.state,
                    )

        if storage_task is None:
            await self._start_task(full_id, task)

    async def _load_previous_run(self) -> None:
        log.debug(f"BatchExecutor: loading previous run")
        # Loads tasks that previous executor run processed.
        # Do not handles fail fast option.

        root_id = ()
        task_id = self._progress.add_task(
            f"<{self._bake.batch}>",
            completed=0,
            total=self._graphs.get_size(root_id) * 3,
        )
        self._bars[root_id] = task_id

        tasks = {task.yaml_id: task async for task in self._storage.list_tasks()}

        def _set_task_info(task: StorageTask) -> None:
            self._update_graph_coloring(task)
            self._advance_progress_bar(None, task)
            self._tasks_mgr.update(task)

        # Rebuild data in graph and task_mgr
        while tasks.keys() != self._tasks_mgr.tasks.keys():
            log.debug(f"BatchExecutor: rebuilding data...")
            for full_id, ctx in self._graphs.get_ready_with_meta():
                if full_id not in tasks or full_id in self._tasks_mgr.tasks:
                    continue
                task = tasks[full_id]
                if await ctx.is_action(full_id[-1]):
                    await self._embed_action(full_id)
                    if task.status.is_pending:
                        _set_task_info(task)
                else:
                    _set_task_info(task)
            for full_id in self._graphs.get_ready_to_mark_running_embeds():
                task = tasks[full_id]
                if task.status.is_finished:
                    self._graphs.mark_running(full_id)
                elif task.status.is_running:
                    _set_task_info(task)
            for full_id in self._graphs.get_ready_to_mark_done_embeds():
                task = tasks[full_id]
                if task.status.is_finished:
                    _set_task_info(task)

    async def _should_continue(self) -> bool:
        return self._graphs.is_active

    async def run(self) -> TaskStatus:
        with self._progress:
            try:
                result = await self._run()
            except (KeyboardInterrupt, asyncio.CancelledError):
                await self._cancel_unfinished()
                result = TaskStatus.CANCELLED
            except Exception as exp:
                await self.log_error(exp)
                await self._cancel_unfinished()
                result = TaskStatus.FAILED
        await self._finish_run(result)
        return result

    async def log_error(self, exp: Exception) -> None:
        if isinstance(exp, EvalError):
            self._progress.log(f"[red][b]ERROR:[/b] {exp}[/red]")
        else:
            # Looks like this is some bug in our code, so we should print stacktrace
            # for easier debug
            self._progress.console.print_exception(show_locals=True)
            self._progress.log(
                "[red][b]ERROR:[/b] Some unknown error happened. Please report "
                "an issue to https://github.com/neuro-inc/neuro-flow/issues/new "
                "with traceback printed above.[/red]"
            )

    async def _run(self) -> TaskStatus:
        await self._load_previous_run()

        job_id = os.environ.get("NEURO_JOB_ID")
        if job_id:
            await self._storage.update(
                executor_id=job_id,
            )

        if self._attempt.result.is_finished:
            # If attempt is already terminated, just clean up tasks
            # and exit
            await self._cancel_unfinished()
            return self._attempt.result

        await self._storage.update(
            result=TaskStatus.RUNNING,
        )

        while await self._should_continue():
            _mgr = self._tasks_mgr

            def _fmt_debug(task: StorageTask) -> str:
                return ".".join(task.yaml_id) + f" - {task.status}"

            log.debug(
                f"BatchExecutor: main loop start:\n"
                f"tasks={', '.join(_fmt_debug(it) for it in _mgr.tasks.values())}\n"
            )
            for full_id, flow in self._graphs.get_ready_with_meta():
                tid = full_id[-1]
                if full_id in self._tasks_mgr.tasks:
                    continue  # Already started
                meta = await self._get_meta(full_id)
                if not meta.enable or (
                    self._is_cancelling and meta.enable is not AlwaysT()
                ):
                    # Make task started and immediatelly skipped
                    await self._skip_task(full_id)
                    continue
                if await flow.is_local(tid):
                    await self._process_local(full_id)
                elif await flow.is_action(tid):
                    await self._process_action(full_id)
                else:
                    await self._process_task(full_id)

            ok = await self._process_started()
            await self._process_running_builds()

            # Check for cancellation
            if not self._is_cancelling:
                await self._refresh_attempt()

                if not ok or self._attempt.result == TaskStatus.CANCELLED:
                    await self._cancel_unfinished()

            await asyncio.sleep(self._polling_timeout or 0)

        return self._accumulate_result()

    async def _finish_run(self, attempt_status: TaskStatus) -> None:
        if not self._attempt.result.is_finished:
            await self._storage.update(result=attempt_status)
        self._progress.print(
            Panel(
                f"[b]Attempt #{self._attempt.number}[/b] {fmt_status(attempt_status)}"
            ),
            justify="center",
        )
        if attempt_status != TaskStatus.SUCCEEDED:
            self._progress.log(
                "[blue b]Hint:[/blue b] you can restart bake starting "
                "from first failed task "
                "by the following command:\n"
                f"[b]apolo-flow restart {self._bake.id}[/b]"
            )

    async def _cancel_unfinished(self) -> None:
        self._is_cancelling = True
        for task, raw_id in self._tasks_mgr.list_unfinished_raw_tasks():
            task_ctx = await self._get_task(task.yaml_id)
            if task_ctx.enable is not AlwaysT():
                self._progress.log(f"Task {fmt_id(task.yaml_id)} is being killed")
                await self._client.job_kill(raw_id)
        async for image in self._bake_storage.list_bake_images():
            if image.status == ImageStatus.BUILDING:
                assert image.builder_job_id
                await self._client.job_kill(image.builder_job_id)

    async def _store_to_cache(self, task: StorageTask) -> None:
        log.debug(f"BatchExecutor: storing to cache {task.yaml_id}")
        task_ctx = await self._get_task(task.yaml_id)
        cache_strategy = task_ctx.cache.strategy
        if (
            cache_strategy == ast.CacheStrategy.NONE
            or task.status != TaskStatus.SUCCEEDED
            or task.raw_id is None
            or task.outputs is None
            or task.state is None
        ):
            return

        await self._project_storage.create_cache_entry(
            key=task_ctx.caching_key,
            batch=self._bake.batch,
            task_id=task.yaml_id,
            raw_id=task.raw_id,
            outputs=task.outputs,
            state=task.state,
        )

    async def _process_started(self) -> bool:
        log.debug(f"BatchExecutor: processing started")
        # Process tasks
        for task, raw_id in self._tasks_mgr.list_unfinished_raw_tasks():
            assert task.raw_id, "list_unfinished_raw_tasks should list raw tasks"
            job_descr = await self._client.job_status(task.raw_id)
            log.debug(
                f"BatchExecutor: got description {job_descr} for task {task.yaml_id}"
            )
            if (
                job_descr.status in {JobStatus.RUNNING, JobStatus.SUCCEEDED}
                and task.status.is_pending
            ):
                assert (
                    job_descr.history.started_at
                ), "RUNNING jobs should have started_at"
                task = await self._update_task(
                    task.yaml_id,
                    new_status=TaskStatusItem(
                        when=job_descr.history.started_at,
                        status=TaskStatus.RUNNING,
                    ),
                )
            if job_descr.status in TERMINATED_JOB_STATUSES:
                log.debug(f"BatchExecutor: processing logs for task {task.yaml_id}")
                async with CmdProcessor() as proc:
                    async for chunk in self._client.job_logs(raw_id):
                        async for line in proc.feed_chunk(chunk):
                            pass
                    async for line in proc.feed_eof():
                        pass
                log.debug(
                    f"BatchExecutor: finished processing logs for task {task.yaml_id}"
                )
                assert (
                    job_descr.history.finished_at
                ), "TERMINATED jobs should have finished_at"
                task = await self._update_task(
                    task.yaml_id,
                    new_status=TaskStatusItem(
                        when=job_descr.history.finished_at,
                        status=TaskStatus(job_descr.status),
                    ),
                    outputs=proc.outputs,
                    state=proc.states,
                )
                await self._store_to_cache(task)
                task_meta = await self._get_meta(task.yaml_id)
                if task.status != TaskStatus.SUCCEEDED and task_meta.strategy.fail_fast:
                    return False
        # Process batch actions
        for full_id in self._graphs.get_ready_to_mark_running_embeds():
            log.debug(f"BatchExecutor: marking action {full_id} as ready")
            await self._update_task(full_id, new_status=TaskStatus.RUNNING)

        for full_id in self._graphs.get_ready_to_mark_done_embeds():
            log.debug(f"BatchExecutor: marking action {full_id} as done")
            # done action, make it finished
            ctx = await self._get_action(full_id)
            results = self._tasks_mgr.build_needs(full_id, ctx.graph.keys())
            res_ctx = await ctx.calc_outputs(results)

            task = await self._update_task(
                full_id,
                new_status=TaskStatus(res_ctx.result),
                outputs=res_ctx.outputs,
                state={},
            )

            task_id = self._bars[full_id]
            self._progress.remove_task(task_id)

            task_meta = await self._get_meta(full_id)
            if task.status != TaskStatus.SUCCEEDED and task_meta.strategy.fail_fast:
                return False
        return True

    def _accumulate_result(self) -> TaskStatus:
        for task in self._tasks_mgr.tasks.values():
            if task.status == TaskStatus.CANCELLED:
                return TaskStatus.CANCELLED
            elif task.status == TaskStatus.FAILED:
                return TaskStatus.FAILED

        return TaskStatus.SUCCEEDED

    async def _is_image_in_registry(self, remote_image: RemoteImage) -> bool:
        try:
            await self._client.image_tag_info(remote_image)
        except ResourceNotFound:
            return False
        return True

    async def _start_task(self, full_id: FullID, task: Task) -> Optional[StorageTask]:
        log.debug(f"BatchExecutor: checking should we build image for {full_id}")
        remote_image = self._client.parse_remote_image(task.image)
        log.debug(f"BatchExecutor: image name is {remote_image}")
        if remote_image.cluster_name is None:  # Not a platform registry image
            return await self._run_task(full_id, task)

        image_storage = self._bake_storage.bake_image(ref=task.image)
        try:
            bake_image = await image_storage.get()
        except ResourceNotFound:
            # Not defined in the bake
            return await self._run_task(full_id, task)

        image_ctx = await self._get_image(bake_image)
        if image_ctx is None:
            # The image referred before definition
            return await self._run_task(full_id, task)

        if not image_ctx.force_rebuild and await self._is_image_in_registry(
            remote_image
        ):
            if bake_image.status == ImageStatus.PENDING:
                await image_storage.update(
                    status=ImageStatus.CACHED,
                )
            return await self._run_task(full_id, task)

        if bake_image.status == ImageStatus.PENDING:
            await self._start_image_build(bake_image, image_ctx, image_storage)
            return None  # wait for next pulling interval
        elif bake_image.status == ImageStatus.BUILDING:
            return None  # wait for next pulling interval
        else:
            # Image is already build (maybe with an error, just try to run a job)
            return await self._run_task(full_id, task)

    async def _run_task(self, full_id: FullID, task: Task) -> StorageTask:
        log.debug(f"BatchExecutor: starting job for {full_id}")
        preset_name = task.preset
        if preset_name is None:
            preset_name = next(iter(self._client.config_presets))

        envs = self._client.parse_envs([f"{k}={v}" for k, v in task.env.items()])

        volumes_parsed = self._client.parse_volumes(task.volumes)
        volumes = list(volumes_parsed.volumes)

        http_auth = task.http_auth
        if http_auth is None:
            http_auth = HTTPPort.requires_auth

        restart_policy = JobRestartPolicy.NEVER
        if task.restart:
            restart_policy = JobRestartPolicy(task.restart)

        job = await self._client.job_start(
            shm=True,
            tty=False,
            image=self._client.parse_remote_image(task.image),
            preset_name=preset_name,
            entrypoint=task.entrypoint,
            command=task.cmd,
            http=HTTPPort(task.http_port, http_auth) if task.http_port else None,
            env=envs.env,
            secret_env=envs.secret_env,
            volumes=volumes,
            working_dir=str(task.workdir) if task.workdir else None,
            secret_files=list(volumes_parsed.secret_files),
            disk_volumes=list(volumes_parsed.disk_volumes),
            name=task.name,
            tags=list(task.tags),
            description=task.title,
            life_span=task.life_span,
            schedule_timeout=task.schedule_timeout,
            pass_config=bool(task.pass_config),
            restart_policy=restart_policy,
            project_name=self._top_flow.project_name,
        )
        return await self._create_task(
            yaml_id=full_id,
            raw_id=job.id,
            status=TaskStatusItem(
                when=job.history.created_at or datetime.now(timezone.utc),
                status=TaskStatus.PENDING,
            ),
        )

    async def _start_image_build(
        self,
        bake_image: BakeImage,
        image_ctx: ImageCtx,
        image_storage: BakeImageStorage,
    ) -> None:
        log.debug(f"BatchExecutor: starting image build for {bake_image}")
        context = bake_image.context_on_storage or image_ctx.context
        dockerfile_rel = bake_image.dockerfile_rel or image_ctx.dockerfile_rel

        if context is None or dockerfile_rel is None:
            if context is None and dockerfile_rel is None:
                error = "context and dockerfile not specified"
            elif context is None and dockerfile_rel is not None:
                error = "context not specified"
            else:
                error = "dockerfile not specified"
            raise Exception(f"Failed to build image '{bake_image.ref}': {error}")

        cmd = ["apolo-extras", "image", "build"]
        cmd.append(f"--project={self._top_flow.project_name}")
        cmd.append(f"--file={dockerfile_rel}")
        for arg in image_ctx.build_args:
            cmd.append(f"--build-arg={arg}")
        for vol in image_ctx.volumes:
            cmd.append(f"--volume={vol}")
        for k, v in image_ctx.env.items():
            cmd.append(f"--env={k}={v}")
        if image_ctx.force_rebuild:
            cmd.append("--force-overwrite")
        if image_ctx.build_preset is not None:
            cmd.append(f"--preset={image_ctx.build_preset}")
        if image_ctx.extra_kaniko_args is not None:
            cmd.append(
                f"--extra-kaniko-args={shlex.quote(image_ctx.extra_kaniko_args)}"
            )
        cmd.append(str(context))
        cmd.append(str(bake_image.ref))
        builder_job_id = await self._run_builder_job(*cmd)
        await image_storage.update(
            builder_job_id=builder_job_id,
            status=ImageStatus.BUILDING,
        )
        self._progress.log("Image", fmt_id(bake_image.ref), "is", ImageStatus.BUILDING)
        self._progress.log("  builder job:", fmt_raw_id(builder_job_id))

    async def _process_running_builds(self) -> None:
        log.debug(f"BatchExecutor: checking running build")
        async for image in self._bake_storage.list_bake_images():
            log.debug(f"BatchExecutor: processing image {image}")
            image_storage = self._bake_storage.bake_image(id=image.id)
            if image.status == ImageStatus.BUILDING:
                assert image.builder_job_id
                descr = await self._client.job_status(image.builder_job_id)
                log.debug(f"BatchExecutor: got job description {descr}")
                if descr.status == JobStatus.SUCCEEDED:
                    await image_storage.update(
                        status=ImageStatus.BUILT,
                    )
                    self._progress.log(
                        "Image", fmt_id(image.ref), "is", ImageStatus.BUILT
                    )
                elif descr.status in TERMINATED_JOB_STATUSES:
                    await image_storage.update(
                        status=ImageStatus.BUILD_FAILED,
                    )
                    self._progress.log(
                        "Image", fmt_id(image.ref), "is", ImageStatus.BUILD_FAILED
                    )


class LocalsBatchExecutor(BatchExecutor):
    transient_progress = True

    @classmethod
    @asynccontextmanager
    async def create(
        cls,
        console: Console,
        bake_id: str,
        client: Client,
        storage: Storage,
        *,
        polling_timeout: Optional[float] = None,
        project_role: Optional[str] = None,
    ) -> AsyncIterator["BatchExecutor"]:
        assert (
            polling_timeout is None
        ), "polling_timeout is disabled for LocalsBatchExecutor"
        async with super(cls, LocalsBatchExecutor).create(
            console,
            bake_id,
            client,
            storage,
            polling_timeout=0,
            project_role=project_role,
        ) as ret:
            yield ret

    async def _process_local(self, full_id: FullID) -> None:
        local = await self._get_local(full_id)
        task = await self._create_task(
            yaml_id=full_id,
            status=TaskStatus.PENDING,
            raw_id=None,
        )
        posix = False if sys.platform == "win32" else True
        subprocess = await asyncio.create_subprocess_exec(
            *shlex.split(local.cmd, posix=posix),
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE,
            cwd=self._top_flow.workspace,
        )
        (stdout_data, stderr_data) = await subprocess.communicate()
        async with CmdProcessor() as proc:
            async for line in proc.feed_chunk(stdout_data):
                self._progress.log(line.decode())
            async for line in proc.feed_eof():
                self._progress.log(line.decode())
        if stderr_data:
            self._progress.log(stderr_data.decode())
        if subprocess.returncode == 0:
            result_status = TaskStatus.SUCCEEDED
        else:
            result_status = TaskStatus.FAILED
        task = await self._update_task(
            task.yaml_id,
            new_status=result_status,
            outputs=proc.outputs,
            state={},
        )

    async def _process_task(self, full_id: FullID) -> None:
        pass  # Skip for local

    async def _should_continue(self) -> bool:
        for full_id, flow in self._graphs.get_ready_with_meta():
            if await flow.is_local(full_id[-1]):
                return True  # Has local action
            if (
                await flow.is_action(full_id[-1])
                and full_id not in self._tasks_mgr.tasks
            ):
                return True  # Has unchecked batch action
        return False

    async def _finish_run(self, attempt_status: TaskStatus) -> None:
        # Only process failures during local run
        if attempt_status != TaskStatus.SUCCEEDED:
            await super()._finish_run(attempt_status)



from dataclasses import replace

import datetime
import secrets
from apolo_sdk import ResourceNotFound
from typing import (
    AbstractSet,
    AsyncIterator,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Type,
    Union,
)
from yarl import URL

from apolo_flow.storage.base import (
    Attempt,
    AttemptStorage,
    Bake,
    BakeImage,
    BakeImageStorage,
    BakeMeta,
    BakeStorage,
    CacheEntry,
    CacheEntryStorage,
    ConfigFile,
    ConfigFileStorage,
    ConfigsMeta,
    LiveJob,
    LiveJobStorage,
    Project,
    ProjectStorage,
    Storage,
    Task,
    TaskStatusItem,
    TaskStorage,
    _Unset,
)
from apolo_flow.types import FullID, ImageStatus, TaskStatus


def _make_id() -> str:
    return secrets.token_hex(10)


def _now() -> datetime.datetime:
    return datetime.datetime.now(datetime.timezone.utc)


class InMemoryDB:
    def __init__(self) -> None:
        self.projects: Dict[str, Project] = {}
        self.bakes: Dict[str, Bake] = {}
        self.cache_entries: Dict[str, CacheEntry] = {}
        self.live_jobs: Dict[str, LiveJob] = {}
        self.attempts: Dict[str, Attempt] = {}
        self.config_files: Dict[str, ConfigFile] = {}
        self.bake_images: Dict[str, BakeImage] = {}
        self.tasks: Dict[str, Task] = {}


class InMemoryStorage(Storage):
    def __init__(
        self,
        owner: str = "in_memory_owner",
        cluster: str = "in_memory_cluster",
        org_name: str = "in_memory_org",
        project_name: str = "in_memory_project",
    ):
        self._owner = owner
        self._cluster = cluster
        self._org_name = org_name
        self._project_name = project_name
        self._db = InMemoryDB()

    def with_retry_read(self) -> "Storage":
        return self  # In memory read cannot fail

    async def close(self) -> None:
        pass

    def project(
        self,
        *,
        id: Optional[str] = None,
        project_name: Optional[str] = None,
        yaml_id: Optional[str] = None,
        cluster: Optional[str] = None,
        org_name: Optional[str] = None,
        owner: Optional[str] = None,
    ) -> "ProjectStorage":
        if id:
            return InMemoryProjectStorage(id, self._db)
        cluster = cluster or self._cluster
        org_name = org_name or self._org_name
        project_name = project_name or self._project_name
        for project in self._db.projects.values():
            if (
                project.yaml_id == yaml_id
                and project.cluster == cluster
                and project.org_name == org_name
                and (owner is None or project.owner == owner)
                and project.project_name == project_name
            ):
                return InMemoryProjectStorage(project.id, self._db)
        return InMemoryProjectStorage("fake-id", self._db)

    def bake(self, *, id: str) -> "BakeStorage":
        return InMemoryBakeStorage(id, self._db)

    async def create_project(
        self,
        yaml_id: str,
        project_name: str,
        cluster: Optional[str] = None,
        org_name: Optional[str] = None,
    ) -> Project:
        project = Project(
            id=_make_id(),
            project_name=project_name,
            yaml_id=yaml_id,
            cluster=cluster or self._cluster,
            owner=self._owner,
            org_name=org_name or self._org_name,
        )
        self._db.projects[project.id] = project
        return project

    async def list_projects(
        self,
        name: Optional[str] = None,
        cluster: Optional[str] = None,
        project_name: Optional[str] = None,
        org_name: Optional[str] = None,
    ) -> AsyncIterator[Project]:
        for project in self._db.projects.values():
            if name and project.yaml_id != name:
                continue
            if cluster and project.cluster != cluster:
                continue
            if project_name and project.project_name != project_name:
                continue
            if org_name and project.org_name != org_name:
                continue
            yield project


class InMemoryProjectStorage(ProjectStorage):
    def __init__(self, project_id: str, db: InMemoryDB) -> None:
        self._project_id = project_id
        self._db = db

    async def get(self) -> Project:
        try:
            return self._db.projects[self._project_id]
        except KeyError:
            raise ResourceNotFound

    async def delete(self) -> None:
        self._db.projects.pop(self._project_id)

    async def list_bakes(
        self,
        tags: Optional[AbstractSet[str]] = None,
        since: Optional[datetime.datetime] = None,
        until: Optional[datetime.datetime] = None,
        recent_first: bool = False,
    ) -> AsyncIterator[Bake]:
        unsorted: List[Bake] = []
        for bake in self._db.bakes.values():
            if bake.project_id != self._project_id:
                continue
            if tags is not None and not set(tags).issubset(set(bake.tags)):
                continue
            if since is not None and bake.created_at <= since:
                continue
            if until is not None and bake.created_at >= until:
                continue
            last_attempt: Optional[Attempt] = None
            try:
                last_attempt = await self.bake(id=bake.id).last_attempt().get()
            except ResourceNotFound:
                pass
            unsorted.append(replace(bake, last_attempt=last_attempt))
        res = sorted(unsorted, key=lambda it: it.created_at)
        if recent_first:
            res = list(reversed(res))
        for item in res:
            yield item

    async def create_bake(
        self,
        batch: str,
        meta: BakeMeta,
        graphs: Mapping[FullID, Mapping[FullID, AbstractSet[FullID]]],
        params: Optional[Mapping[str, str]] = None,
        name: Optional[str] = None,
        tags: Sequence[str] = (),
    ) -> Bake:
        bake = Bake(
            id=_make_id(),
            project_id=self._project_id,
            batch=batch,
            graphs=graphs,
            params=params,
            name=name,
            tags=tags,
            created_at=_now(),
            last_attempt=None,
            meta=meta,
        )
        self._db.bakes[bake.id] = bake
        return bake

    def bake(
        self, *, id: Optional[str] = None, name: Optional[str] = None
    ) -> "BakeStorage":
        if id:
            return InMemoryBakeStorage(id, self._db)
        else:
            bakes: List[Bake] = []
            for bake in self._db.bakes.values():
                if bake.name == name and bake.project_id == self._project_id:
                    bakes.append(bake)
            bakes.sort(key=lambda it: it.created_at)
            if bakes:
                return InMemoryBakeStorage(bakes[-1].id, self._db)
        return InMemoryBakeStorage("fake-id", self._db)  # Error should be lazy

    async def create_cache_entry(
        self,
        task_id: FullID,
        batch: str,
        key: str,
        outputs: Mapping[str, str],
        state: Mapping[str, str],
        raw_id: str,
    ) -> CacheEntry:
        entry = CacheEntry(
            id=_make_id(),
            project_id=self._project_id,
            batch=batch,
            task_id=task_id,
            key=key,
            created_at=_now(),
            outputs=outputs,
            state=state,
            raw_id=raw_id,
        )
        self._db.cache_entries[entry.id] = entry
        return entry

    async def delete_cache_entries(
        self, batch: Optional[str] = None, task_id: Optional[FullID] = None
    ) -> None:
        if batch is None:
            self._db.cache_entries.clear()
        for entry in list(self._db.cache_entries.values()):
            if entry.batch == batch and (task_id is None or entry.task_id == task_id):
                self._db.cache_entries.pop(entry.id)

    def cache_entry(
        self,
        *,
        id: Optional[str] = None,
        task_id: Optional[FullID] = None,
        batch: Optional[str] = None,
        key: Optional[str] = None,
    ) -> "CacheEntryStorage":
        if id:
            return InMemoryCacheEntryStorage(id, self._db)
        for entry in self._db.cache_entries.values():
            if (
                entry.project_id == self._project_id
                and entry.task_id == task_id
                and entry.batch == batch
                and entry.key == key
            ):
                return InMemoryCacheEntryStorage(entry.id, self._db)
        return InMemoryCacheEntryStorage("fake-id", self._db)  # Error should be lazy

    async def list_live_jobs(self) -> AsyncIterator[LiveJob]:
        for live_job in self._db.live_jobs.values():
            if live_job.project_id == self._project_id:
                yield live_job

    async def create_live_job(
        self,
        yaml_id: str,
        multi: bool,
        tags: Iterable[str],
        raw_id: Optional[str] = None,
    ) -> LiveJob:
        live_job = LiveJob(
            id=_make_id(),
            project_id=self._project_id,
            yaml_id=yaml_id,
            multi=multi,
            tags=list(tags),
            raw_id=raw_id,
        )
        self._db.live_jobs[live_job.id] = live_job
        return live_job

    async def replace_live_job(
        self,
        yaml_id: str,
        multi: bool,
        tags: Iterable[str],
        raw_id: Optional[str] = None,
    ) -> LiveJob:
        try:
            live_job_id = (await self.live_job(yaml_id=yaml_id).get()).id
        except ResourceNotFound:
            return await self.create_live_job(yaml_id, multi, tags, raw_id)
        else:
            live_job = LiveJob(
                id=live_job_id,
                project_id=self._project_id,
                yaml_id=yaml_id,
                multi=multi,
                tags=list(tags),
                raw_id=raw_id,
            )
            self._db.live_jobs[live_job_id] = live_job
            return live_job

    def live_job(
        self, *, id: Optional[str] = None, yaml_id: Optional[str] = None
    ) -> "LiveJobStorage":
        if id:
            return InMemoryLiveJobStorage(id, self._db)
        else:
            for live_job in self._db.live_jobs.values():
                if (
                    live_job.yaml_id == yaml_id
                    and live_job.project_id == self._project_id
                ):
                    return InMemoryLiveJobStorage(live_job.id, self._db)
        return InMemoryLiveJobStorage("fake-id", self._db)  # Error should be lazy


class InMemoryBakeStorage(BakeStorage):
    def __init__(self, bake_id: str, db: InMemoryDB) -> None:
        self._bake_id = bake_id
        self._db = db

    async def get(self) -> Bake:
        try:
            return self._db.bakes[self._bake_id]
        except KeyError:
            raise ResourceNotFound

    async def list_attempts(self) -> AsyncIterator[Attempt]:
        for attempt in self._db.attempts.values():
            if attempt.bake_id == self._bake_id:
                yield attempt

    async def create_attempt(
        self,
        configs_meta: ConfigsMeta,
        number: Optional[int] = None,
        executor_id: Optional[str] = None,
        result: TaskStatus = TaskStatus.PENDING,
    ) -> Attempt:
        attempts_cnt = len([_ async for _ in self.list_attempts()])
        attempt = Attempt(
            id=_make_id(),
            bake_id=self._bake_id,
            configs_meta=configs_meta,
            number=number or attempts_cnt + 1,
            executor_id=executor_id,
            result=result,
            created_at=_now(),
        )
        self._db.attempts[attempt.id] = attempt
        return attempt

    async def create_config_file(self, filename: str, content: str) -> ConfigFile:
        config_file = ConfigFile(
            id=_make_id(),
            bake_id=self._bake_id,
            filename=filename,
            content=content,
        )
        self._db.config_files[config_file.id] = config_file
        return config_file

    async def list_bake_images(self) -> AsyncIterator[BakeImage]:
        for bake_image in self._db.bake_images.values():
            if bake_image.bake_id == self._bake_id:
                yield bake_image

    async def create_bake_image(
        self,
        yaml_defs: Sequence[FullID],
        ref: str,
        status: ImageStatus = ImageStatus.PENDING,
        context_on_storage: Optional[URL] = None,
        dockerfile_rel: Optional[str] = None,
        builder_job_id: Optional[str] = None,
    ) -> BakeImage:
        bake_image = BakeImage(
            id=_make_id(),
            ref=ref,
            bake_id=self._bake_id,
            yaml_defs=yaml_defs,
            status=status,
            context_on_storage=context_on_storage,
            dockerfile_rel=dockerfile_rel,
            builder_job_id=builder_job_id,
        )
        self._db.bake_images[bake_image.id] = bake_image
        return bake_image

    def attempt(
        self, *, id: Optional[str] = None, number: Optional[int] = None
    ) -> "AttemptStorage":
        if id:
            return InMemoryAttemptStorage(id, self._db)
        else:
            if number == -1:
                for attempt in self._db.attempts.values():
                    if attempt.bake_id == self._bake_id:
                        number = max(number, attempt.number)
            for attempt in self._db.attempts.values():
                if attempt.number == number and attempt.bake_id == self._bake_id:
                    return InMemoryAttemptStorage(attempt.id, self._db)
        return InMemoryAttemptStorage("fake-id", self._db)  # Error should be lazy

    def config_file(self, *, id: str) -> "ConfigFileStorage":
        return InMemoryConfigFileStorage(id, self._db)

    def bake_image(
        self, *, id: Optional[str] = None, ref: Optional[str] = None
    ) -> "BakeImageStorage":
        if id:
            return InMemoryBakeImageStorage(id, self._db)
        else:
            for image in self._db.bake_images.values():
                if image.ref == ref and image.bake_id == self._bake_id:
                    return InMemoryBakeImageStorage(image.id, self._db)
        return InMemoryBakeImageStorage("fake-id", self._db)  # Error should be lazy


class InMemoryAttemptStorage(AttemptStorage):
    def __init__(self, attempt_id: str, db: InMemoryDB):
        self._db = db
        self._attempt_id = attempt_id

    async def get(self) -> Attempt:
        try:
            return self._db.attempts[self._attempt_id]
        except KeyError:
            raise ResourceNotFound

    async def update(
        self,
        *,
        executor_id: Union[Optional[str], Type[_Unset]] = _Unset,
        result: Union[TaskStatus, Type[_Unset]] = _Unset,
    ) -> Attempt:
        attempt = await self.get()
        if executor_id is not _Unset:
            attempt = replace(
                attempt, executor_id=executor_id  # type: ignore[arg-type]
            )
        if result is not _Unset:
            attempt = replace(attempt, result=result)  # type: ignore[arg-type]
        self._db.attempts[attempt.id] = attempt
        return attempt

    async def list_tasks(self) -> AsyncIterator[Task]:
        for task in self._db.tasks.values():
            if task.attempt_id == self._attempt_id:
                yield task

    async def create_task(
        self,
        yaml_id: FullID,
        raw_id: Optional[str],
        status: Union[TaskStatusItem, TaskStatus],
        outputs: Optional[Mapping[str, str]] = None,
        state: Optional[Mapping[str, str]] = None,
    ) -> Task:
        if isinstance(status, TaskStatus):
            status = TaskStatusItem(when=_now(), status=status)
        task = Task(
            id=_make_id(),
            attempt_id=self._attempt_id,
            yaml_id=yaml_id,
            raw_id=raw_id,
            outputs=outputs,
            state=state,
            statuses=[status],
        )
        self._db.tasks[task.id] = task
        return task

    def task(
        self, *, id: Optional[str] = None, yaml_id: Optional[FullID] = None
    ) -> "TaskStorage":
        if id:
            return InMemoryTaskStorage(id, self._db)
        else:
            for task in self._db.tasks.values():
                if task.yaml_id == yaml_id and task.attempt_id == self._attempt_id:
                    return InMemoryTaskStorage(task.id, self._db)
        return InMemoryTaskStorage("fake-id", self._db)  # Error should be lazy


class InMemoryTaskStorage(TaskStorage):
    def __init__(self, task_id: str, db: InMemoryDB):
        self._db = db
        self._task_id = task_id

    async def get(self) -> Task:
        try:
            return self._db.tasks[self._task_id]
        except KeyError:
            raise ResourceNotFound

    async def update(
        self,
        *,
        outputs: Union[Optional[Mapping[str, str]], Type[_Unset]] = _Unset,
        state: Union[Optional[Mapping[str, str]], Type[_Unset]] = _Unset,
        new_status: Optional[Union[TaskStatusItem, TaskStatus]] = None,
    ) -> Task:
        task = await self.get()
        if outputs is not _Unset:
            task = replace(task, outputs=outputs)  # type: ignore[arg-type]
        if state is not _Unset:
            task = replace(task, state=state)  # type: ignore[arg-type]
        if isinstance(new_status, TaskStatus):
            new_status = TaskStatusItem(when=_now(), status=new_status)
        if new_status:
            task = replace(task, statuses=[*task.statuses, new_status])
        self._db.tasks[task.id] = task
        return task


class InMemoryConfigFileStorage(ConfigFileStorage):
    def __init__(self, config_file_id: str, db: InMemoryDB):
        self._db = db
        self._config_file_id = config_file_id

    async def get(self) -> ConfigFile:
        try:
            return self._db.config_files[self._config_file_id]
        except KeyError:
            raise ResourceNotFound


class InMemoryBakeImageStorage(BakeImageStorage):
    def __init__(self, bake_image_id: str, db: InMemoryDB):
        self._db = db
        self._bake_image_id = bake_image_id

    async def get(self) -> BakeImage:
        try:
            return self._db.bake_images[self._bake_image_id]
        except KeyError:
            raise ResourceNotFound

    async def update(
        self,
        *,
        status: Union[ImageStatus, Type[_Unset]] = _Unset,
        builder_job_id: Union[Optional[str], Type[_Unset]] = _Unset,
    ) -> BakeImage:
        image = await self.get()
        if status is not _Unset:
            image = replace(image, status=status)  # type: ignore[arg-type]
        if builder_job_id is not _Unset:
            image = replace(
                image, builder_job_id=builder_job_id  # type: ignore[arg-type]
            )
        self._db.bake_images[image.id] = image
        return image


class InMemoryCacheEntryStorage(CacheEntryStorage):
    def __init__(self, cache_entry_id: str, db: InMemoryDB):
        self._db = db
        self._cache_entry_id = cache_entry_id

    async def get(self) -> CacheEntry:
        try:
            return self._db.cache_entries[self._cache_entry_id]
        except KeyError:
            raise ResourceNotFound


class InMemoryLiveJobStorage(LiveJobStorage):
    def __init__(self, live_job_id: str, db: InMemoryDB):
        self._db = db
        self._live_job_id = live_job_id

    async def get(self) -> LiveJob:
        try:
            return self._db.live_jobs[self._live_job_id]
        except KeyError:
            raise ResourceNotFound
